# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-arguments

"""PeriFlow GroupClient Service."""

from __future__ import annotations

import json
import uuid
from string import Template
from typing import Any, Dict, List, Optional
from uuid import UUID

from periflow.client.base import (
    Client,
    GroupRequestMixin,
    ProjectRequestMixin,
    UserRequestMixin,
    safe_request,
)
from periflow.enums import CheckpointCategory, StorageType
from periflow.utils.request import paginated_get
from periflow.utils.url import get_auth_uri, get_mr_uri


class GroupClient(Client):
    """Organization client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_group"))

    def create_group(self, name: str) -> Dict[str, Any]:
        """Create a new organization."""
        response = safe_request(
            self.post, err_prefix="Failed to post an organization."
        )(data=json.dumps({"name": name, "hosting_type": "hosted"}))
        return response.json()

    def get_group(self, pf_group_id: uuid.UUID) -> Dict[str, Any]:
        """Get the organization info."""
        response = safe_request(
            self.retrieve, err_prefix="Failed to get an organization."
        )(pk=pf_group_id)
        return response.json()

    def invite_to_group(self, pf_group_id: uuid.UUID, email: str) -> None:
        """Invite a new member to the organization by sending an email."""
        safe_request(self.post, err_prefix="Failed to send invitation")(
            path=f"{pf_group_id}/invite/signup",
            json={
                "email": email,
                "callback_path": "/api/auth/invite/callback",
            },
        )

    def list_users(self, pf_group_id: uuid.UUID) -> List[Dict[str, Any]]:
        """List all organization member info."""
        get_response_dict = safe_request(
            self.list, err_prefix="Failed to list users in organization"
        )
        return paginated_get(get_response_dict, path=f"{pf_group_id}/pf_user")


class GroupProjectClient(Client, GroupRequestMixin):
    """Organization project client."""

    def __init__(self, **kwargs):
        """Initialize group project client."""
        self.initialize_group()
        super().__init__(pf_group_id=self.group_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_group/$pf_group_id/pf_project"))

    def create_project(self, name: str) -> Dict[str, Any]:
        """Create a new project in the organization."""
        response = safe_request(self.post, err_prefix="Failed to post a project.")(
            data=json.dumps({"name": name})
        )
        return response.json()

    def list_projects(self) -> List[Dict[str, Any]]:
        """List all projects in the organization."""
        get_response_dict = safe_request(
            self.list, err_prefix="Failed to list projects."
        )
        return paginated_get(get_response_dict)


class GroupProjectCheckpointClient(
    Client, UserRequestMixin, GroupRequestMixin, ProjectRequestMixin
):
    """Organization project checkpoint client."""

    def __init__(self, **kwargs):
        """Initialize organization project checkpoint client."""
        self.initialize_user()
        self.initialize_group()
        self.initialize_project()
        super().__init__(group_id=self.group_id, project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_mr_uri("orgs/$group_id/prjs/$project_id/models/"))

    def list_checkpoints(
        self, category: Optional[CheckpointCategory], limit: int, deleted: bool
    ) -> List[Dict[str, Any]]:
        """List checkpoints."""
        request_data = {}
        if category is not None:
            request_data["category"] = category.value
        if deleted:
            request_data["status"] = "deleted"

        get_response_dict = safe_request(
            self.list, err_prefix="Failed to list checkpoints."
        )
        return paginated_get(get_response_dict, **request_data, limit=limit)

    def create_checkpoint(
        self,
        name: str,
        vendor: StorageType,
        region: str,
        credential_id: Optional[UUID],
        iteration: Optional[int],
        storage_name: str,
        files: List[Dict[str, Any]],
        dist_config: Dict[str, Any],
        attributes: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Create a new checkpoint."""
        request_data = {
            "job_id": None,
            "name": name,
            "attributes": attributes,
            "user_id": str(self.user_id),
            "secret_type": "credential",
            "secret_id": str(credential_id) if credential_id else None,
            "model_category": "USER",
            "form_category": "ORCA",
            "dist_json": dist_config,
            "vendor": vendor,
            "region": region,
            "storage_name": storage_name,
            "iteration": iteration,
            "files": files,
        }

        response = safe_request(self.post, err_prefix="Failed to post checkpoint.")(
            json=request_data
        )
        return response.json()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow SDK errors."""

from __future__ import annotations

from typing import Any, List, Optional


class PeriFlowError(Exception):
    """PeriFlow exception base."""


class AuthenticationError(PeriFlowError):
    """Authentication failure error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthTokenNotFoundError."""
        super().__init__(f"Failed to authenticate: {detail}")


class AuthorizationError(PeriFlowError):
    """Not authorized to access."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthorizationError."""
        super().__init__(f"Not authorized to access: {detail}")


class AuthTokenNotFoundError(PeriFlowError):
    """Auth token is not found."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize AuthTokenNotFoundError."""
        super().__init__(f"Auth token is not found: {detail}")


class LowServicePlanError(PeriFlowError):
    """The feature is not supported in your service plan."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize LowServicePlanError."""
        super().__init__(f"The feature is not supported in your service plan: {detail}")


class InvalidConfigError(PeriFlowError):
    """Invalid configuration provided."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidConfigError."""
        super().__init__(f"Invalid configuration provided: {detail}")


class InvalidAttributesError(PeriFlowError):
    """Checkpoint attributes are not valid."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidAttributesError."""
        super().__init__(f"Invalid checkpoint attributes: {detail}")


class InvalidPathError(PeriFlowError):
    """Invalid path provided."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize InvalidPathError."""
        super().__init__(f"Invalid path provided: {detail}")


class EntityTooLargeError(PeriFlowError):
    """Entity is too large."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize EntityTooLargeError."""
        super().__init__(f"Entity is too large: {detail}")


class NotFoundError(PeriFlowError):
    """Requested resource is not found."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotFoundError."""
        super().__init__(f"The resource is not found: {detail}")


class CheckpointConversionError(PeriFlowError):
    """Checkpoint conversion failure error."""

    def __init__(self, msg: str) -> None:
        """Initialize CheckpointConversionError."""
        super().__init__(f"Conversion failed: {msg}")


class NotSupportedCheckpointError(CheckpointConversionError):
    """Checkpoint is not supported."""

    def __init__(self, invalid_option: str, valid_options: List[Any]) -> None:
        """Initialize NotSupportedCheckpointError."""
        super().__init__(
            f"{invalid_option} is not supported. Please use one of {valid_options}."
        )


class TokenizerNotFoundError(PeriFlowError):
    """Cannot find PeriFlow-compatible tokenizer info."""

    def __init__(self, msg: Optional[str] = None) -> None:
        """Initialize TokenizerNotFoundError."""
        docs_link = (
            "https://huggingface.co/docs/transformers/main_classes/tokenizer#tokenizer"
        )
        default_msg = (
            "PeriFlow only supports Hugging Face 'fast' tokenizer. "
            f"Refer to {docs_link} to get more info."
        )
        super().__init__(msg or default_msg)


class RequestTimeoutError(PeriFlowError):
    """Request timeout error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotFoundError."""
        super().__init__(f"Request timeout: {detail}")


class APIError(PeriFlowError):
    """PeriFlow API error."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize APIError."""
        super().__init__(f"API error: {detail}")


class SessionClosedError(PeriFlowError):
    """API Session is closed."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize SessionClosedError."""
        super().__init__(
            f"Cannot send requests to the API server because session is not opened: {detail}"
        )


class NotSupportedError(PeriFlowError):
    """Feature is not supported."""

    def __init__(self, detail: Optional[str] = None) -> None:
        """Initialize NotSupportedError."""
        super().__init__(f"The feature is not supported: {detail}")


class PeriFlowInternalError(PeriFlowError):
    """Internal error of PeriFlow."""

    def __init__(self, message: Optional[str] = None) -> None:
        """Initialize PeriFlowInternalError."""
        if message is None:
            message = "Please contact to FriendliAI"
        else:
            message = message.rstrip() + "\nPlease contant to FriendliAI"

        super().__init__(message)


class InvalidGenerationError(PeriFlowInternalError):
    """Invalid generation error."""

    def __init__(self, message: Optional[str] = None) -> None:
        """Initialize invalid generation error."""
        super().__init__(f"Invalid generation result: {message}")



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-locals, redefined-builtin, too-many-statements, too-many-arguments

"""PeriFlow Deployment CLI."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional
from uuid import UUID

import typer
import yaml
from dateutil.parser import parse

from periflow.client.user import UserGroupProjectClient
from periflow.enums import CloudType, DeploymentSecurityLevel, DeploymentType, VMType
from periflow.errors import (
    AuthenticationError,
    EntityTooLargeError,
    InvalidConfigError,
    LowServicePlanError,
)
from periflow.formatter import PanelFormatter, TableFormatter
from periflow.sdk.resource.deployment import Deployment as DeploymentAPI
from periflow.utils.format import (
    datetime_to_pretty_str,
    datetime_to_simple_string,
    secho_error_and_exit,
)

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

DEFAULT_MAX_BATCH_SIZE = 256
DEFAULT_MAX_TOKEN_COUNT = 8192

deployment_panel = PanelFormatter(
    name="Deployment Overview",
    fields=[
        "deployment_id",
        "config.name",
        "description",
        "config.deployment_type",
        "config.model_id",
        "status",
        "ready_replicas",
        "config.scaler_config.min_replica_count",
        "config.scaler_config.max_replica_count",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "end",
        "security_level",
        "config.infrequest_log",
        "endpoint",
        "config.cloud",
        "config.region",
        "config.orca_config.max_batch_size",
        "config.orca_config.max_token_count",
    ],
    headers=[
        "ID",
        "Name",
        "Description",
        "Type",
        "Ckpt ID",
        "Status",
        "#Ready",
        "Min Replicas",
        "Max Replicas",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "End",
        "Security Level",
        "Logging",
        "Endpoint",
        "Cloud",
        "Region",
        "Max batch size",
        "Max token count",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_table = TableFormatter(
    name="Deployments",
    fields=[
        "deployment_id",
        "config.name",
        "status",
        "ready_replicas",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "config.cloud",
        "config.region",
    ],
    headers=[
        "ID",
        "Name",
        "Status",
        "#Ready",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "Cloud",
        "Region",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_org_table = TableFormatter(
    name="Deployments",
    fields=[
        "deployment_id",
        "config.name",
        "status",
        "ready_replicas",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "config.cloud",
        "config.region",
        "config.project_id",
        "project_name",
    ],
    headers=[
        "ID",
        "Name",
        "Status",
        "#Ready",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "Cloud",
        "Region",
        "Project ID",
        "Project Name",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_metrics_table = TableFormatter(
    name="Deployment Metrics",
    fields=[
        "id",
        "latency",
        "throughput",
        "time_window",
    ],
    headers=["ID", "Latency(ms)", "Throughput(req/s)", "Time Window(sec)"],
    extra_fields=["error"],
    extra_headers=["error"],
)

deployment_usage_table = TableFormatter(
    name="Deployment Usage",
    fields=[
        "id",
        "type",
        "cloud",
        "vm",
        "created_at",
        "finished_at",
        "gpu_type",
        "duration",
    ],
    headers=[
        "ID",
        "Type",
        "Cloud",
        "VM",
        "Created At",
        "Finished At",
        "GPU",
        "Total Usage (days, HH:MM:SS)",
    ],
)

deployment_event_table = TableFormatter(
    name="Deployment Event",
    fields=[
        "id",
        "type",
        "description",
        "created_at",
    ],
    headers=["ID", "Type", "Description", "Timestamp"],
)

deployment_panel.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_panel.add_substitution_rule("Healthy", "[bold green]Healthy[/bold green]")
deployment_panel.add_substitution_rule("Unhealthy", "[bold red]Unhealthy[/bold red]")
deployment_panel.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_panel.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")

deployment_table.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_table.add_substitution_rule("Healthy", "[bold green]Healthy[/bold green]")
deployment_table.add_substitution_rule("Unhealthy", "[bold red]Unhealthy[/bold red]")
deployment_table.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_table.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")

deployment_org_table.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_org_table.add_substitution_rule(
    "Healthy", "[bold green]Healthy[/bold green]"
)
deployment_org_table.add_substitution_rule(
    "Unhealthy", "[bold red]Unhealthy[/bold red]"
)
deployment_org_table.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_org_table.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")


def get_deployment_id_from_namespace(namespace: str):
    """Get deployment id from namespace."""
    return f"periflow-deployment-{namespace}"


@app.command()
def list(
    include_terminated: bool = typer.Option(
        False,
        "--include-terminated",
        help=(
            "When True, shows all deployments in my project including terminated "
            "deployments. (active deployments are shown above the terminated ones.)"
        ),
    ),
    limit: int = typer.Option(20, "--limit", help="The number of deployments to view"),
    from_oldest: bool = typer.Option(
        False, "--from-oldest", help="When True, shows from the oldest deployments."
    ),
    org: bool = typer.Option(False, "--org", help="Shows all deployments in org"),
):
    """Lists all deployments."""
    try:
        deployments = DeploymentAPI.list(
            limit=limit,
            include_terminated=include_terminated,
            from_oldest=from_oldest,
            all_in_org=org,
        )
    except AuthenticationError as exc:
        secho_error_and_exit(str(exc))

    deployment_dicts = []
    for deployment in deployments:
        deployment_dict = deployment.model_dump()
        deployment_dict["start"] = datetime_to_pretty_str(deployment.start)
        deployment_dict["vms"] = deployment.vms[0].name if deployment.vms else "None"
        deployment_dicts.append(deployment_dict)

    table = deployment_table
    if org:
        table = deployment_org_table
        project_client = UserGroupProjectClient()
        projects = project_client.list_projects()
        project_map = {project["id"]: project["name"] for project in projects}
        for deployment_dict in deployment_dicts:
            project_id = str(deployment_dict["config"]["project_id"])
            deployment_dict["project_name"] = (
                project_map[project_id] if project_id in project_map else project_id
            )

    table.render(deployment_dicts)


@app.command()
def stop(deployment_id: str = typer.Argument(..., help="ID of deployment to stop")):
    """Stops a running deployment."""
    DeploymentAPI.stop(id=deployment_id)
    typer.secho(
        f"Deployment ({deployment_id}) stopped successfully.",
        fg=typer.colors.GREEN,
    )


@app.command()
def view(
    deployment_id: str = typer.Argument(..., help="deployment id to inspect detail.")
):
    """Shows details of a deployment."""
    deployment = DeploymentAPI.get(id=deployment_id)

    deployment_dict = deployment.model_dump()
    deployment_dict["start"] = deployment.start.ctime()
    deployment_dict["end"] = deployment.end and deployment.end.ctime()
    deployment_dict["vms"] = deployment.vms[0].name if deployment.vms else None
    deployment_dict["security_level"] = (
        DeploymentSecurityLevel.PROTECTED
        if deployment.config.infrequest_perm_check
        else DeploymentSecurityLevel.PUBLIC
    )
    deployment_panel.render([deployment_dict])


@app.command()
def metrics(
    deployment_id: str = typer.Argument(
        ..., help="ID of deployment to inspect in detail."
    ),
    time_window: int = typer.Option(
        60, "--time-window", "-t", help="Time window of metrics in seconds."
    ),
):
    """Show metrics of a deployment."""
    metrics = DeploymentAPI.get_metrics(id=deployment_id, time_window=time_window)
    metrics["id"] = metrics["deployment_id"]
    if metrics["latency"]:
        # ns => ms
        metrics["latency"] = (
            f"{metrics['latency'] / 1000000:.3f}" if "latency" in metrics else None
        )
    if metrics["throughput"]:
        metrics["throughput"] = (
            f"{metrics['throughput']:.3f}" if "throughput" in metrics else None
        )
    deployment_metrics_table.render([metrics])


@app.command()
def usage(
    year: int = typer.Argument(...),
    month: int = typer.Argument(...),
    day: Optional[int] = typer.Argument(None),
):
    """Shows the usage of all deployments in a project within a specific month or day.

    :::info
    - Timestamps are recorded in UTC.
    - The usages are updated every minute.
    :::

    """
    try:
        start_date = datetime(year, month, day if day else 1, tzinfo=timezone.utc)
    except ValueError:
        secho_error_and_exit(f"Invalid date({year}-{month}{f'-{day}' if day else ''})")
    if day:
        end_date = start_date + timedelta(days=1)
    else:
        end_date = datetime(
            year + int(month == 12),
            (month + 1) if month < 12 else 1,
            1,
            tzinfo=timezone.utc,
        )
    usages = DeploymentAPI.get_usage(start_date, end_date)
    deployments = [
        {
            "id": id,
            "type": info["deployment_type"],
            "cloud": info["cloud"].upper() if "cloud" in info else None,
            "vm": info["vm"]["name"] if info.get("vm") else None,
            "gpu_type": info["vm"]["gpu_type"].upper() if info.get("vm") else None,
            "created_at": datetime_to_simple_string(parse(info["created_at"])),
            "finished_at": datetime_to_simple_string(parse(info["finished_at"]))
            if info["finished_at"]
            else "-",
            "duration": timedelta(seconds=int(info["duration"])),
        }
        for id, info in usages.items()
        if int(info["duration"]) != 0
    ]
    deployment_usage_table.render(deployments)


@app.command()
def log(
    deployment_id: str = typer.Argument(..., help="ID of deployment to get log."),
    replica_index: int = typer.Argument(
        0, help="Replica index of deployment to get log."
    ),
):
    """Shows deployments log."""
    logs = DeploymentAPI.get_logs(id=deployment_id, replica_index=replica_index)
    for line in logs:
        typer.echo(line["data"])


@app.command()
def create(
    checkpoint_id: UUID = typer.Option(
        ..., "--checkpoint-id", "-i", help="Checkpoint id to deploy."
    ),
    deployment_name: str = typer.Option(
        ..., "--name", "-n", help="The name of deployment. "
    ),
    cloud: CloudType = typer.Option(..., "--cloud", "-c", help="Type of cloud."),
    region: str = typer.Option(..., "--region", "-r", help="Region of cloud."),
    vm_type: VMType = typer.Option(
        ..., "--vm-type", "-v", help="The VM type for the deployment."
    ),
    config_file: Optional[typer.FileText] = typer.Option(
        None, "--config-file", "-f", help="Path to configuration file."
    ),
    deployment_type: DeploymentType = typer.Option(
        DeploymentType.PROD, "--type", "-t", help="Type of deployment."
    ),
    description: Optional[str] = typer.Option(
        None, "--description", "-d", help="Deployment description."
    ),
    security_level: DeploymentSecurityLevel = typer.Option(
        DeploymentSecurityLevel.PUBLIC.value,
        "--security-level",
        "-sl",
        help="Security level of deployment endpoints",
    ),
    logging: bool = typer.Option(
        False,
        "--logging",
        "-l",
        help="Logging inference requests or not.",
    ),
    default_request_config_file: Optional[typer.FileText] = typer.Option(
        None,
        "--default-request-config-file",
        "-drc",
        help="Path to default request config",
    ),
    min_replicas: int = typer.Option(
        1,
        "--min-replicas",
        "-min",
        help="Number of minimum replicas.",
    ),
    max_replicas: int = typer.Option(
        1,
        "--max-replicas",
        "-max",
        help="Number of maximum replicas.",
    ),
):
    """Creates a deployment object by using model checkpoint.

    The deployment settings are described in a configuration YAML file, and the path of that file is
    passed to the `-f` option. The following is an example YAML file:

    ```yaml
    orca_config:
        max_batch_size: 384
        max_token_count: 12288
    ```

    :::tip
    To turn off the deployment autoscaling, set `--min-replicas` and
    `--max-replicas` to the same value.
    :::

    :::tip
    Use `pf vm list` to find available vm-type, cloud, region, and gpu-type.
    :::

    The default request-response configuration, such as stop tokens or bad words, is
    defined in a YAML file. The path of that file is passed to the `-drc` option.
    The format of the file is as follows.

    ```json
    {
        "stop": Optional[array[str]],
        "stop_tokens": Optional[array[int]],
        "bad_words": Optional[array[str]],
        "bad_word_tokens": Optional[array[int]]
    }
    ```

    :::caution
    Both `bad_words` and `bad_word_tokens` cannot be set at the same time. Also, both
    `stop` and `stop_tokens` cannot be set at the same time.
    :::

    """
    default_request_config = None
    config: Dict[str, Any] = {}
    if config_file:
        try:
            config = yaml.safe_load(config_file)
            if default_request_config_file is not None:
                default_request_config = yaml.safe_load(default_request_config_file)
        except yaml.YAMLError as e:
            secho_error_and_exit(
                f"Error occurred while parsing engine config file... {e}"
            )
    else:
        config["orca_config"] = {
            "max_batch_size": DEFAULT_MAX_BATCH_SIZE,
            "max_token_count": DEFAULT_MAX_TOKEN_COUNT,
        }

    try:
        deployment = DeploymentAPI.create(
            checkpoint_id=checkpoint_id,
            name=deployment_name,
            deployment_type=deployment_type,
            cloud=cloud,
            region=region,
            vm_type=vm_type,
            config=config,
            description=description,
            default_request_config=default_request_config,
            security_level=security_level,
            logging=logging,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )
    except (
        AuthenticationError,
        InvalidConfigError,
        EntityTooLargeError,
        LowServicePlanError,
    ) as exc:
        secho_error_and_exit(str(exc))

    typer.secho(
        f"Deployment ({deployment.deployment_id}) started successfully. "
        f"Use 'pf deployment view {deployment.deployment_id}' to see the deployment details.\n"
        f"Send inference requests to '{deployment.endpoint}'.",
        fg=typer.colors.GREEN,
    )


@app.command()
def update(
    deployment_id: str = typer.Argument(..., help="ID of deployment to update."),
    min_replicas: int = typer.Option(
        ..., "--min-replicas", "-min", help="Set min_replicas of deployment."
    ),
    max_replicas: int = typer.Option(
        ..., "--max-replicas", "-max", help="Set max_replicas of deployment."
    ),
):
    """Updates configuration of a deployment.

    :::tip
    To turn off the deployment autoscaling, set `--min-replicas` and
    `--max-replicas` to the same value.
    :::

    """
    DeploymentAPI.adjust_replica_config(
        id=deployment_id, min_replicas=min_replicas, max_replicas=max_replicas
    )
    typer.secho(
        f"Scaler of deployment ({deployment_id}) is updated.\n"
        f"Set min_replicas to {min_replicas}, max_replicas to {max_replicas}",
        fg=typer.colors.GREEN,
    )


@app.command()
def event(
    deployment_id: str = typer.Argument(..., help="ID of deployment to get events."),
):
    """Gets deployment events."""
    events = DeploymentAPI.get_events(id=deployment_id)
    for event in events:
        event["id"] = f"periflow-deployment-{event['namespace']}"
        event["created_at"] = datetime_to_simple_string(parse(event["created_at"]))
    deployment_event_table.render(events)


@app.command()
def req_resp(
    deployment_id: str = typer.Argument(
        ..., help="ID of deployment to download request-response logs."
    ),
    since: str = typer.Option(
        ...,
        "--since",
        help=(
            "Start time of request-response logs. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    until: str = typer.Option(
        ...,
        "--until",
        help=(
            "End time of request-response logs. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    save_dir: Optional[str] = typer.Option(
        None,
        "-s",
        "--save-dir",
        help="Directory path to save request-response logs",
    ),
):
    """Downloads request-response logs for a deployment."""
    try:
        start = datetime.strptime(since, "%Y-%m-%dT%H")
        end = datetime.strptime(until, "%Y-%m-%dT%H")
    except ValueError:
        secho_error_and_exit(
            "Invalid datetime format. The format should be {YYYY}-{MM}-{DD}T{HH} "
            "(e.g., 1999-01-01T01)."
        )

    DeploymentAPI.download_req_resp_logs(
        id=deployment_id, since=start, until=end, save_dir=save_dir
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI Validation Utilities."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Type

import typer
from pydantic import ValidationError

from periflow.cloud.storage import cloud_region_map, storage_region_map
from periflow.enums import CloudType, StorageType
from periflow.errors import (
    InvalidAttributesError,
    InvalidConfigError,
    NotSupportedError,
)
from periflow.schema.resource.v1.attributes import V1AttributesValidationModel
from periflow.utils.format import secho_error_and_exit
from periflow.utils.version import (
    PERIFLOW_PACKAGE_NAME,
    get_installed_version,
    get_latest_version,
    is_latest_version,
)


def validate_storage_region(vendor: StorageType, region: str) -> None:
    """Validation the cloud storage availability region."""
    available_regions = storage_region_map[vendor]
    if region not in available_regions:
        raise InvalidConfigError(
            f"'{region}' is not supported region for {vendor}. "
            f"Please choose another one in {available_regions}."
        )


def validate_cloud_region(vendor: CloudType, region: str) -> None:
    """Validate the cloud availability region."""
    available_regions = cloud_region_map[vendor]
    if region not in available_regions:
        secho_error_and_exit(
            f"'{region}' is not a supported region for {vendor}. "
            f"Please choose another one in {available_regions}."
        )


def validate_datetime_format(datetime_str: Optional[str]) -> Optional[str]:
    """Validate the datetime format."""
    if datetime_str is None:
        return None

    try:
        datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S")
    except ValueError as exc:
        raise typer.BadParameter(
            "The datetime format should be {YYYY}-{MM}-{DD}T{HH}:{MM}:{SS}"
        ) from exc
    return datetime_str


def validate_cloud_storage_type(val: StorageType) -> None:
    """Validate the cloud storage type."""
    if val == StorageType.FAI:
        raise NotSupportedError(
            "Checkpoint creation with FAI storage is not supported now."
        )


def validate_package_version() -> None:
    """Validate the installed CLI version."""
    installed_version = get_installed_version()
    if not is_latest_version(installed_version):
        latest_version = get_latest_version()
        secho_error_and_exit(
            f"Package version({installed_version}) is deprecated. "
            f"Please install the latest version({latest_version}) with "
            f"'pip install {PERIFLOW_PACKAGE_NAME}=={latest_version} -U'."
        )


def validate_checkpoint_attributes(attr: Dict[str, Any]) -> None:
    """Validate checkpoint attributes schema."""
    try:
        V1AttributesValidationModel.model_validate({"attr": attr})
    except ValidationError as exc:
        msgs = []
        for error in exc.errors():
            error_type = error["type"]
            if error_type == "union_tag_invalid":
                msgs.append("'model_type' is not one of the expected values.")
            elif error_type == "union_tag_not_found":
                msgs.append("'model_type' filed is missing.")
            else:
                msgs.append(f"{error['msg']}. Correct the field '{error['loc'][-1]}'")

        if len(msgs) == 1:
            msg = msgs[0]
        else:
            msg = "\n>>> " + "\n>>> ".join(msgs)
        raise InvalidAttributesError(msg) from exc


def validate_enums(val: Any, enum_cls: Type[Enum]) -> Any:
    """Validate if the value is the proper enum."""
    try:
        return enum_cls(val)
    except ValueError as exc:
        supported_values = set([e.value for e in enum_cls])
        raise InvalidConfigError(
            f"Invalid value. Please provide one of {supported_values}"
        ) from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI."""

# pylint: disable=line-too-long

from __future__ import annotations

import requests
import typer
from requests import HTTPError, Response

from periflow.auth import TokenType, clear_tokens, get_token, update_token
from periflow.cli import checkpoint, credential, deployment, group, key, project, vm
from periflow.client.project import ProjectClient
from periflow.client.user import UserClient, UserGroupClient, UserMFAClient
from periflow.context import (
    get_current_project_id,
    project_context_path,
    set_current_group_id,
)
from periflow.formatter import PanelFormatter
from periflow.utils.format import secho_error_and_exit
from periflow.utils.request import DEFAULT_REQ_TIMEOUT
from periflow.utils.url import get_training_uri
from periflow.utils.validate import validate_package_version
from periflow.utils.version import get_installed_version

app = typer.Typer(
    help="Supercharge Generative AI Serving 🚀",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
    callback=validate_package_version,
    pretty_exceptions_enable=False,
)

app.add_typer(credential.app, name="credential", help="Manage credentials")
app.add_typer(checkpoint.app, name="checkpoint", help="Manage checkpoints")
app.add_typer(vm.app, name="vm", help="Manage VMs")
app.add_typer(deployment.app, name="deployment", help="Manage deployments")
app.add_typer(project.app, name="project", help="Manage projects")
app.add_typer(group.app, name="org", help="Manage organizations")
app.add_typer(key.app, name="key", help="Manage api keys")


user_panel_formatter = PanelFormatter(
    name="My Info",
    fields=["name", "email"],
    headers=["Name", "Email"],
)


@app.command()
def whoami():
    """Show my user info."""
    client = UserClient()
    info = client.get_current_userinfo()
    user_panel_formatter.render([info])


@app.command()
def login(
    email: str = typer.Option(..., prompt="Enter your email"),
    password: str = typer.Option(..., prompt="Enter your password", hide_input=True),
):
    """Sign in."""
    r = requests.post(
        get_training_uri("token/"),
        data={"username": email, "password": password},
        timeout=DEFAULT_REQ_TIMEOUT,
    )
    resp = r.json()
    if "code" in resp and resp["code"] == "mfa_required":
        mfa_token = resp["mfa_token"]
        client = UserMFAClient()
        # TODO: MFA type currently defaults to totp, need changes when new options are added
        client.initiate_mfa(mfa_type="totp", mfa_token=mfa_token)
        update_token(token_type=TokenType.MFA, token=mfa_token)
        typer.run(_mfa_verify)
    else:
        _handle_login_response(r, False)

    # Save user's organiztion context
    project_client = ProjectClient()
    user_group_client = UserGroupClient()

    try:
        org = user_group_client.get_group_info()
    except IndexError:
        secho_error_and_exit("You are not included in any organization.")
    org_id = org["id"]

    project_id = get_current_project_id()
    if project_id is not None:
        if project_client.check_project_membership(pf_project_id=project_id):
            project_org_id = project_client.get_project(pf_project_id=project_id)[
                "pf_group_id"
            ]
            if project_org_id != org_id:
                project_context_path.unlink(missing_ok=True)
        else:
            project_context_path.unlink(missing_ok=True)
    set_current_group_id(org_id)


@app.command()
def logout():
    """Sign out."""
    clear_tokens()
    typer.secho("Successfully signed out.", fg=typer.colors.BLUE)


@app.command()
def passwd(
    old_password: str = typer.Option(
        ..., prompt="Enter your current password", hide_input=True
    ),
    new_password: str = typer.Option(
        ..., prompt="Enter your new password", hide_input=True
    ),
    confirm_password: str = typer.Option(
        ..., prompt="Enter the new password again (confirmation)", hide_input=True
    ),
):
    """Change password."""
    if old_password == new_password:
        secho_error_and_exit("The current password is the same with the new password.")
    if new_password != confirm_password:
        secho_error_and_exit("Passwords did not match.")
    client = UserClient()
    client.change_password(old_password, new_password)

    typer.secho("Password is changed successfully!", fg=typer.colors.BLUE)


@app.command()
def version():
    """Check the installed package version."""
    installed_version = get_installed_version()
    typer.echo(installed_version)


def _mfa_verify(_, code: str = typer.Option(..., prompt="Enter MFA Code")):
    mfa_token = get_token(TokenType.MFA)
    # TODO: MFA type currently defaults to totp, need changes when new options are added
    mfa_type = "totp"
    username = f"mfa://{mfa_type}/{mfa_token}"
    r = requests.post(
        get_training_uri("token/"),
        data={"username": username, "password": code},
        timeout=DEFAULT_REQ_TIMEOUT,
    )
    _handle_login_response(r, True)


def _handle_login_response(r: Response, mfa: bool):
    try:
        r.raise_for_status()
        update_token(token_type=TokenType.ACCESS, token=r.json()["access_token"])
        update_token(token_type=TokenType.REFRESH, token=r.json()["refresh_token"])

        typer.echo("\n\nLogin success!")
        typer.echo("Welcome back to...")
        typer.echo(r" _____          _  _____ _")
        typer.echo(r"|  __ \___ _ __(_)|  ___| | _____      __")  # type: ignore
        typer.echo(r"|  ___/ _ \ '__| || |__ | |/ _ \ \ /\ / /")  # type: ignore
        typer.echo(r"| |  |  __/ |  | ||  __|| | (_) | V  V / ")
        typer.echo(r"|_|   \___|_|  |_||_|   |_|\___/ \_/\_/  ")  # type: ignore
        typer.echo("\n\n")
    except HTTPError:
        if mfa:
            secho_error_and_exit("Login failed... Invalid MFA Code.")
        else:
            secho_error_and_exit(
                "Login failed... Please check your email and password."
            )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI URL Utilities."""

from __future__ import annotations

from urllib.parse import urljoin, urlparse

training_url = "https://training.periflow.ai/api/"
training_ws_url = "wss://training-ws.periflow.ai/ws/"
discuss_url = "https://discuss.friendli.ai/"
registry_url = "https://modelregistry.periflow.ai/"
serving_url = "https://serving.periflow.ai/"
auth_url = "https://auth.periflow.ai/"
meter_url = "https://metering.periflow.ai/"
observatory_url = "https://observatory.periflow.ai/"


def get_auth_uri(path: str) -> str:
    """Get PFA URI."""
    return urljoin(auth_url, path)


def get_training_uri(path: str) -> str:
    """Get PFT URI."""
    return urljoin(training_url, path)


def get_training_ws_uri(path: str) -> str:
    """Get PFT websocket URI."""
    return urljoin(training_ws_url, path)


def get_serving_uri(path: str) -> str:
    """Get PFS URI."""
    return urljoin(serving_url, path)


def get_mr_uri(path: str) -> str:
    """Get PFR URI."""
    return urljoin(registry_url, path)


def get_meter_uri(path: str) -> str:
    """Get PFM URI."""
    return urljoin(meter_url, path)


def get_observatory_uri(path: str) -> str:
    """Get PFO URI."""
    return urljoin(observatory_url, path)


def get_baseurl(url: str) -> str:
    """Get a base of a URL."""
    parsed_url = urlparse(url)
    scheme = parsed_url.scheme
    base = parsed_url.netloc
    return f"{scheme}://{base}/"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow UserClient Service."""

from __future__ import annotations

from string import Template
from typing import Any, Dict, List
from uuid import UUID

from periflow.client.base import (
    Client,
    GroupRequestMixin,
    UserRequestMixin,
    safe_request,
)
from periflow.enums import GroupRole, ProjectRole
from periflow.utils.request import paginated_get
from periflow.utils.url import get_auth_uri


class UserMFAClient(Client):
    """User MFA client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("mfa"))

    def initiate_mfa(self, mfa_type: str, mfa_token: str) -> None:
        """Authenticate by MFA token."""
        safe_request(self.bare_post, err_prefix="Failed to verify MFA token.")(
            path=f"challenge/{mfa_type}", headers={"x-mfa-token": mfa_token}
        )


class UserSignUpClient(Client):
    """User sign-up client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_user/self_signup"))

    def verify(self, token: str, key: str) -> None:
        """Verify the email account with the token to sign up."""
        safe_request(self.bare_post, err_prefix="Failed to verify")(
            path="confirm", json={"email_token": token, "key": key}
        )


class UserClient(Client, UserRequestMixin):
    """User client."""

    def __init__(self, **kwargs) -> None:
        """Initialize user client."""
        self.initialize_user()
        super().__init__(**kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_user"))

    def change_password(self, old_password: str, new_password: str) -> None:
        """Change password."""
        safe_request(self.update, err_prefix="Failed to change password.")(
            pk=self.user_id,
            path="password",
            json={"old_password": old_password, "new_password": new_password},
        )

    def set_group_privilege(
        self, pf_group_id: UUID, pf_user_id: UUID, privilege_level: GroupRole
    ) -> None:
        """Update user role in the orgnaization."""
        safe_request(
            self.partial_update, err_prefix="Failed to update privilege level in group"
        )(
            pk=pf_user_id,
            path=f"pf_group/{pf_group_id}/privilege_level",
            json={"privilege_level": privilege_level.value},
        )

    def get_project_membership(self, pf_project_id: UUID) -> Dict[str, Any]:
        """Get the project membership info of the user."""
        response = safe_request(
            self.retrieve, err_prefix="Failed identify member in project"
        )(
            pk=self.user_id,
            path=f"pf_project/{pf_project_id}",
        )
        return response.json()

    def add_to_project(
        self, pf_user_id: UUID, pf_project_id: UUID, access_level: ProjectRole
    ) -> None:
        """Add a new member to a project."""
        safe_request(self.post, err_prefix="Failed to add user to project")(
            path=f"{pf_user_id}/pf_project/{pf_project_id}",
            json={"access_level": access_level.value},
        )

    def delete_from_org(self, pf_user_id: UUID, pf_org_id: UUID) -> None:
        """Delete a member from the organization."""
        safe_request(self.delete, err_prefix="Failed to remove user from organization")(
            pk=pf_user_id, path=f"pf_group/{pf_org_id}"
        )

    def delete_from_project(self, pf_user_id: UUID, pf_project_id: UUID) -> None:
        """Delete a member from the organization."""
        safe_request(self.delete, err_prefix="Failed to remove user from proejct")(
            pk=pf_user_id, path=f"pf_project/{pf_project_id}"
        )

    def set_project_privilege(
        self, pf_user_id: UUID, pf_project_id: UUID, access_level: ProjectRole
    ) -> None:
        """Set a project-level role to a user."""
        safe_request(
            self.partial_update,
            err_prefix="Failed to update privilege level in project",
        )(
            pk=pf_user_id,
            path=f"pf_project/{pf_project_id}/access_level",
            json={"access_level": access_level.value},
        )


class UserGroupClient(Client, UserRequestMixin):
    """User organization client."""

    def __init__(self, **kwargs):
        """Initialize user organization client."""
        self.initialize_user()
        super().__init__(pf_user_id=self.user_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_user/$pf_user_id/pf_group"))

    def get_group_info(self) -> Dict[str, Any]:
        """Get organization info where user belongs to."""
        response = safe_request(self.list, err_prefix="Failed to get my group info.")()
        return response.json()[0]


class UserGroupProjectClient(Client, UserRequestMixin, GroupRequestMixin):
    """User organization project client."""

    def __init__(self, **kwargs):
        """Initialize user organization project client."""
        self.initialize_user()
        self.initialize_group()
        super().__init__(pf_user_id=self.user_id, pf_group_id=self.group_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            get_auth_uri("pf_user/$pf_user_id/pf_group/$pf_group_id/pf_project")
        )

    def list_projects(self) -> List[Dict[str, Any]]:
        """List projects in the organization."""
        get_response_dict = safe_request(
            self.list, err_prefix="Failed to list projects."
        )
        return paginated_get(get_response_dict)


class UserAccessKeyClient(Client, UserRequestMixin):
    """User access key client."""

    def __init__(self, **kwargs):
        """Initialize user access key client."""
        self.initialize_user()
        super().__init__(pf_user_id=self.user_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_user"))

    def create_access_key(self, name: str) -> Dict[str, Any]:
        """Create a new access key."""
        response = safe_request(
            self.post, err_prefix="Failed to create new access key."
        )(path=f"{self.user_id}/api_key", json={"name": name})

        return response.json()

    def delete_access_key(self, access_key_id: str) -> None:
        """Revoke an access key."""
        safe_request(self.delete, err_prefix="Failed to delete access key")(
            pk=None, path=f"api_key/{access_key_id}"
        )

    def list_access_keys(self) -> List[Dict[str, Any]]:
        """List access keys."""
        response = safe_request(
            self.list, err_prefix="Failed to list available access keys."
        )(path=f"{self.user_id}/api_key")
        return response.json()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Client Service."""

from __future__ import annotations

import copy
import math
import os
import uuid
from abc import ABC, abstractmethod
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from string import Template
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union
from urllib.parse import urljoin, urlparse

import requests
from requests.models import Response
from tqdm import tqdm

from periflow.auth import auto_token_refresh, get_auth_header
from periflow.context import get_current_group_id, get_current_project_id
from periflow.errors import AuthTokenNotFoundError
from periflow.utils.format import secho_error_and_exit
from periflow.utils.fs import (
    S3_MPU_PART_MAX_SIZE,
    get_file_size,
    get_total_file_size,
    storage_path_to_local_path,
    upload_file,
    upload_part,
)
from periflow.utils.request import DEFAULT_REQ_TIMEOUT, decode_http_err
from periflow.utils.url import get_auth_uri

T = TypeVar("T", bound=Union[int, str, uuid.UUID])


def safe_request(
    func: Callable[..., Response], *, err_prefix: str = ""
) -> Callable[..., Response]:
    """Wrapper function to send requests with an error checking."""
    if err_prefix:
        err_prefix = err_prefix.rstrip() + "\n"

    @wraps(func)
    def wrapper(*args, **kwargs) -> Response:  # type: ignore
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as exc:
            # typer.secho(exc.response.content)
            secho_error_and_exit(err_prefix + decode_http_err(exc))

    return wrapper


@dataclass
class URLTemplate:
    """URL template."""

    pattern: Template

    def render(
        self, pk: Optional[T] = None, path: Optional[str] = None, **kwargs
    ) -> str:
        """Render URLTemplate.

        Args:
            pk: Primary key of a resource.
            path: Additional URL path to attach.

        """
        if pk is None and path is None:
            return self.pattern.substitute(**kwargs)

        pattern = copy.deepcopy(self.pattern)
        need_trailing_slash = pattern.template.endswith("/")

        if pk is not None:
            pattern.template = urljoin(pattern.template + "/", str(pk))
            if need_trailing_slash:
                pattern.template += "/"

        if path is not None:
            pattern.template = urljoin(pattern.template + "/", path.rstrip("/"))
            if need_trailing_slash:
                pattern.template += "/"

        return pattern.substitute(**kwargs)

    def get_base_url(self) -> str:
        """Get a base URL."""
        result = urlparse(self.pattern.template)
        return f"{result.scheme}://{result.hostname}"

    def attach_pattern(self, pattern: str) -> None:
        """Attach a URL path pattern."""
        self.pattern.template = urljoin(self.pattern.template + "/", pattern)

    def replace_path(self, path: str):
        """Replace a URL path."""
        result = urlparse(self.pattern.template)
        result = result._replace(path=path)
        self.pattern.template = result.geturl()

    def copy(self) -> "URLTemplate":
        """Get a copy of this URL template."""
        return URLTemplate(pattern=Template(self.pattern.template))


class Client(ABC, Generic[T]):
    """Base interface of client to PeriFlow system."""

    def __init__(self, **kwargs):
        """Initialize client."""
        self.url_template = URLTemplate(self.url_path)
        self.url_kwargs = kwargs

    @property
    @abstractmethod
    def url_path(self) -> Template:
        """URL path template to render."""

    @property
    def default_request_options(self) -> Dict[str, Any]:
        """Common request options."""
        try:
            return {
                "headers": get_auth_header(),
                "timeout": DEFAULT_REQ_TIMEOUT,
            }
        except AuthTokenNotFoundError as exc:
            secho_error_and_exit(str(exc))

    @auto_token_refresh
    def list(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to list objects."""
        return requests.get(
            self.url_template.render(path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    @auto_token_refresh
    def retrieve(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to retrieve a specific object."""
        return requests.get(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    @auto_token_refresh
    def post(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to post an object."""
        return requests.post(
            self.url_template.render(path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    @auto_token_refresh
    def partial_update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to partially update a specific object."""
        return requests.patch(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    @auto_token_refresh
    def delete(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to delete a specific obejct."""
        return requests.delete(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    @auto_token_refresh
    def update(self, pk: T, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request to update a specific object."""
        return requests.put(
            self.url_template.render(pk=pk, path=path, **self.url_kwargs),
            **self.default_request_options,
            **kwargs,
        )

    def bare_post(self, path: Optional[str] = None, **kwargs) -> Response:
        """Send a request without automatic auth."""
        r = requests.post(
            self.url_template.render(path=path, **self.url_kwargs),
            timeout=DEFAULT_REQ_TIMEOUT,
            **kwargs,
        )
        r.raise_for_status()
        return r


class UserRequestMixin:
    """Mixin class of user-level requests."""

    user_id: uuid.UUID

    @auto_token_refresh
    def _userinfo(self) -> Response:
        try:
            return requests.get(
                get_auth_uri("oauth2/userinfo"),
                headers=get_auth_header(),
                timeout=DEFAULT_REQ_TIMEOUT,
            )
        except AuthTokenNotFoundError as exc:
            secho_error_and_exit(str(exc))

    def get_current_userinfo(self) -> Dict[str, Any]:
        """Get the currently logged-in user info."""
        response = safe_request(self._userinfo, err_prefix="Failed to get userinfo.")()
        return response.json()

    def get_current_user_id(self) -> uuid.UUID:
        """Get the currently logged-in user ID."""
        userinfo = self.get_current_userinfo()
        return uuid.UUID(userinfo["sub"].split("|")[1])

    def initialize_user(self):
        """Initialize user settings."""
        self.user_id = self.get_current_user_id()


class GroupRequestMixin:
    """Mixin class of organization-level requests."""

    group_id: uuid.UUID

    def initialize_group(self):
        """Initialize organization settings."""
        group_id = get_current_group_id()
        if group_id is None:
            secho_error_and_exit("Organization is not set.")
        self.group_id = group_id  # type: ignore


class ProjectRequestMixin:
    """Mixin class of project-level requests."""

    project_id: uuid.UUID

    def initialize_project(self):
        """Initialize project settings."""
        project_id = get_current_project_id()
        if project_id is None:
            secho_error_and_exit("Project is not set.")
        self.project_id = project_id  # type: ignore


class UploadableClient(Client[T], Generic[T]):
    """Uploadable client."""

    def get_spu_urls(
        self,
        obj_id: T,
        storage_paths: List[str],
    ) -> List[Dict[str, Any]]:
        """Get single part upload URLs for multiple files.

        Args:
            obj_id (T): Uploadable object ID
            storage_paths (List[str]): A list of cloud storage paths of target files

        Returns:
            List[Dict[str, Any]]: A response body that has presigned URL info to upload files.

        """
        response = safe_request(self.post, err_prefix="Failed to get presigned URLs.")(
            path=f"{obj_id}/upload/", json={"paths": storage_paths}
        )
        return response.json()

    def get_mpu_urls(
        self,
        obj_id: T,
        local_paths: List[str],
        storage_paths: List[str],
    ) -> List[Dict[str, Any]]:
        """Get multipart upload URLs for multiple file-like objects.

        Args:
            obj_id (T): Uploadable object ID
            local_paths (List[str]): A list local paths to target files. The path can be
                either absolute or relative.
            storage_paths (List[str]): A list of storage paths to target files.

        Returns:
            List[Dict[str, Any]]: A list of multipart upload responses for multiple target files.

        """
        start_mpu_resps = []
        for local_path, storage_path in zip(local_paths, storage_paths):
            num_parts = math.ceil(get_file_size(local_path) / S3_MPU_PART_MAX_SIZE)
            response = safe_request(
                self.post,
                err_prefix="Failed to get presigned URLs for multipart upload.",
            )(
                path=f"{obj_id}/start_mpu/",
                json={
                    "path": storage_path,
                    "num_parts": num_parts,
                },
            )
            start_mpu_resps.append(response.json())
        return start_mpu_resps

    def complete_mpu(
        self, obj_id: T, path: str, upload_id: str, parts: List[Dict[str, Any]]
    ) -> None:
        """Complete multipart upload.

        Args:
            obj_id (T): Uploadable object ID
            path (str): Path to the uploaded file
            upload_id (str): Upload ID
            parts (List[Dict[str, Any]]): A list of upload part info

        """
        safe_request(
            self.post, err_prefix=f"Failed to complete multipart upload for {path}"
        )(
            path=f"{obj_id}/complete_mpu/",
            json={
                "path": path,
                "upload_id": upload_id,
                "parts": parts,
            },
        )

    def abort_mpu(self, obj_id: T, path: str, upload_id: str) -> None:
        """Abort multipart upload.

        Args:
            obj_id (T): Uploadable object ID
            path (str): Path to the target file
            upload_id (str): Upload ID
            parts (List[Dict[str, Any]]): A list of upload part info

        """
        safe_request(
            self.post, err_prefix=f"Failed to abort multipart upload for {path}"
        )(
            path=f"{obj_id}/abort_mpu/",
            json={
                "path": path,
                "upload_id": upload_id,
            },
        )

    def _multipart_upload_file(
        self,
        obj_id: T,
        file_path: str,
        url_dict: Dict[str, Any],
        ctx: tqdm,
        executor: ThreadPoolExecutor,
    ) -> None:
        """Start multi-part uploading for a file."""
        parts = []
        upload_id = url_dict["upload_id"]
        object_path = url_dict["path"]
        upload_urls = url_dict["upload_urls"]
        total_num_parts = len(upload_urls)
        try:
            futs = [
                executor.submit(
                    upload_part,
                    file_path=file_path,
                    chunk_index=idx,
                    part_number=url_info["part_number"],
                    upload_url=url_info["upload_url"],
                    ctx=ctx,
                    is_last_part=(idx == total_num_parts - 1),
                )
                for idx, url_info in enumerate(upload_urls)
            ]
            wait(futs, return_when=FIRST_EXCEPTION)
            for fut in futs:
                exc = fut.exception()
                if exc is not None:
                    raise exc
                parts.append(fut.result())
            self.complete_mpu(obj_id, object_path, upload_id, parts)
        except KeyboardInterrupt:
            secho_error_and_exit("File upload is aborted.")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.abort_mpu(obj_id, object_path, upload_id)
            secho_error_and_exit(f"File upload is aborted: ({exc!r})")

    def upload_files(
        self,
        obj_id: T,
        spu_url_dicts: List[Dict[str, str]],
        mpu_url_dicts: List[Dict[str, Any]],
        source_path: Path,
        max_workers: int = min(
            32, (os.cpu_count() or 1) + 4
        ),  # default of ``ThreadPoolExecutor``
    ) -> None:
        """Upload files.

        Args:
            obj_id (T): PeriFlow resource(i.e., checkpoint) object ID.
            spu_url_dicts (List[Dict[str, str]]): Single-part upload URL info.
            mpu_url_dicts (List[Dict[str, Any]]): Multi-part upload URL info.
            source_path (Path): Path to the local source directory.
            max_workers (int, optional): Max number of workers for the concurrent uploading.

        """
        spu_local_paths = [
            storage_path_to_local_path(url_info["path"], source_path)
            for url_info in spu_url_dicts
        ]
        mpu_local_paths = [
            storage_path_to_local_path(url_info["path"], source_path)
            for url_info in mpu_url_dicts
        ]
        total_size = get_total_file_size(spu_local_paths + mpu_local_paths)
        spu_urls = [url_info["upload_url"] for url_info in spu_url_dicts]

        with tqdm(
            total=total_size, unit="B", unit_scale=True, unit_divisor=1024
        ) as pbar:
            # NOTE: excessive concurrency may results in "No buffer space available" error.
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                # Normal upload for files with size < 5 GiB
                futs = [
                    executor.submit(upload_file, local_path, upload_url, pbar)
                    for (local_path, upload_url) in zip(spu_local_paths, spu_urls)
                ]
                # Multipart upload for large files with sizes >= 5 GiB
                futs.extend(
                    [
                        executor.submit(
                            self._multipart_upload_file,
                            obj_id=obj_id,
                            file_path=local_path,
                            url_dict=url_dict,
                            ctx=pbar,
                            executor=executor,
                        )
                        for (local_path, url_dict) in zip(
                            mpu_local_paths, mpu_url_dicts
                        )
                    ]
                )
                wait(futs, return_when=FIRST_EXCEPTION)
                for fut in futs:
                    exc = fut.exception()
                    if exc is not None:
                        raise exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow deployment SDK."""

# pylint: disable=line-too-long, arguments-differ, too-many-arguments, too-many-locals, redefined-builtin

from __future__ import annotations

import os
from datetime import datetime
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional
from uuid import UUID

import yaml
from dateutil.tz import tzlocal
from tqdm import tqdm

from periflow.client.deployment import (
    DeploymentClient,
    DeploymentEventClient,
    DeploymentLogClient,
    DeploymentMetricsClient,
    DeploymentReqRespClient,
    PFSProjectUsageClient,
)
from periflow.client.file import FileClient, GroupProjectFileClient
from periflow.client.group import GroupClient
from periflow.configurator.deployment import DRCConfigurator, OrcaDeploymentConfigurator
from periflow.context import get_current_group_id, get_current_project_id
from periflow.enums import (
    CloudType,
    DeploymentSecurityLevel,
    DeploymentType,
    ServiceTier,
    VMType,
)
from periflow.errors import (
    AuthenticationError,
    EntityTooLargeError,
    InvalidConfigError,
    InvalidPathError,
    LowServicePlanError,
    NotFoundError,
)
from periflow.logging import logger
from periflow.schema.resource.v1.deployment import V1Deployment
from periflow.sdk.resource.base import ResourceAPI
from periflow.utils.format import extract_datetime_part, extract_deployment_id_part
from periflow.utils.fs import download_file, upload_file
from periflow.utils.maps import cloud_vm_map, vm_num_gpu_map
from periflow.utils.validate import validate_enums


class Deployment(ResourceAPI[V1Deployment, str]):
    """Deployment resource API."""

    @staticmethod
    def create(
        checkpoint_id: UUID,
        name: str,
        cloud: CloudType,
        region: str,
        vm_type: VMType,
        config: Dict[str, Any],
        deployment_type: DeploymentType = DeploymentType.PROD,
        description: Optional[str] = None,
        default_request_config: Optional[Dict[str, Any]] = None,
        security_level: DeploymentSecurityLevel = DeploymentSecurityLevel.PUBLIC,
        logging: bool = False,
        min_replicas: int = 1,
        max_replicas: int = 1,
    ) -> V1Deployment:
        """Creates a new deployment.

        Args:
            checkpoint_id (UUID): ID of checkpoint to deploy.
            name (str): The name of deployment.
            cloud (CloudType): Type of cloud provider.
            region (str): Cloud region to create a deployment.
            vm_type (VMType): Type of VM.
            config (Dict[str, Any]): Deployment configuration.
            deployment_type (DeploymentType, optional): Type of deployment. Defaults to DeploymentType.PROD.
            description (Optional[str], optional): Optional long description for the deployment. Defaults to None.
            default_request_config (Optional[Dict[str, Any]], optional): Default request configuration (e.g., stop words, bad words). Defaults to None.
            security_level (DeploymentSecurityLevel, optional): Security level of deployment endpoint. Defaults to DeploymentSecurityLevel.PUBLIC.
            logging (bool, optional): When True, enables request-response logging for the deployment if it is set. Defaults to False.
            min_replicas (int, optional): The number of minimum replicas. Defaults to 1.
            max_replicas (int, optional): The number of maximum replicas. Defaults to 1.

        Raises:
            AuthenticationError: Raised when project context is not configured.
            InvalidConfigError: Raised when any of the configurations (i.e., `config`, `default_request_config`, `min_replicas`, `max_replicas`) is invalid.
            EntityTooLargeError: Raised when the contents of `default_request_config` exceeds the 10GiB size limit.
            LowServicePlanError: Raised when the `deployment_type` is set to `Deployment.DEV` and service plan of the user's organization is not 'enterprise'.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERIFLOW_API_KEY",
                project_name="my-project",
            )

            # Create a deployment at GCP asia-northest3 region wtih one A100 GPU.
            config = {
                "max_batch_size": 256,
                "max_token_count": 8146,
            }
            deployment = pf.Deployment.create(
                checkpoint_id="YOUR_CHECKPOINT_ID",
                name="my-deployment",
                cloud="gcp",
                region="asia-northeast3",
                vm_type="a2-highgpu-1g",
                config=config,
            )
            ```

            The format of `config` should be:

            ```python
            {
                "orca_config": {
                    "max_batch_size": Optioanl[int],
                    "max_token_count": Optioanl[int],
                }
            }
            ```

            The format of `default_request_config` should be:

            ```python
            {
                "stop": Optional[List[str]],
                "stop_tokens": Optional[List[int]],
                "bad_words": Optional[List[str]],
                "bad_word_tokens": Optional[List[int]]
            }
            ```

            :::caution
            Note that `bad_words` and `bad_word_tokens` cannot be set at the same time. Similarly, `stop` and `stop_tokens` cannot be set at the same time.
            :::

            :::note
            When `min_replicas` and `max_replicas` are the same, deployment auto-scaling turns off.
            :::

        Returns:
            V1Deployment: The created deployment object.

        """
        # pylint: disable=too-many-statements
        cloud = validate_enums(cloud, CloudType)
        vm_type = validate_enums(vm_type, VMType)
        deployment_type = validate_enums(deployment_type, DeploymentType)
        security_level = validate_enums(security_level, DeploymentSecurityLevel)

        org_id = get_current_group_id()
        if org_id is None:
            raise AuthenticationError(
                "Not authenticated. Please authenticate with 'pf init()' or 'pf login'."
            )
        project_id = get_current_project_id()
        if project_id is None:
            raise AuthenticationError(
                "Project context is not configured. "
                "Set the project context either by 'pf.init()' or 'pf project switch'."
            )

        group_client = GroupClient()
        if (
            group_client.get_group(pf_group_id=org_id)["plan"] == ServiceTier.BASIC
            and deployment_type == DeploymentType.DEV
        ):
            raise LowServicePlanError(
                "Deployment with the development type is only supported for the 'enterprise' plan."
            )

        if min_replicas > max_replicas:
            raise InvalidConfigError(
                f"Should be min_replicas('{min_replicas}') <= max_replicas('{max_replicas}')."
            )

        if vm_type not in cloud_vm_map[cloud]:
            raise InvalidConfigError(
                f"VM type {vm_type.value} is not supported in cloud {cloud.value}."
            )

        deploy_configurator = OrcaDeploymentConfigurator(config=config)
        deploy_configurator.validate()

        if default_request_config is not None:
            drc_configurator = DRCConfigurator(config=default_request_config)
            drc_configurator.validate()

            file_client = FileClient()
            group_file_client = GroupProjectFileClient()

            with TemporaryDirectory() as dir:
                drc_file_name = "drc.yaml"
                drc_file_path = os.path.join(dir, drc_file_name)
                with open(drc_file_path, "w", encoding="utf-8") as file:
                    yaml.dump(default_request_config, file)

                file_size = os.stat(drc_file_path).st_size
                if file_size > 10737418240:  # 10GiB
                    raise EntityTooLargeError(
                        "The default request config size should be smaller than 10GiB."
                    )

                file_info = {
                    "name": drc_file_name,
                    "path": drc_file_name,
                    "mtime": datetime.fromtimestamp(
                        os.stat(drc_file_path).st_mtime, tz=tzlocal()
                    ).isoformat(),
                    "size": file_size,
                }
                file_id = group_file_client.create_misc_file(file_info=file_info)["id"]

                upload_url = file_client.get_misc_file_upload_url(misc_file_id=file_id)
                with tqdm(
                    total=file_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    desc="Uploading default request config",
                ) as t:
                    upload_file(
                        file_path=drc_file_path,
                        url=upload_url,
                        ctx=t,
                    )
                file_client.make_misc_file_uploaded(misc_file_id=file_id)
                config["orca_config"]["default_request_config_id"] = file_id

        num_gpus = vm_num_gpu_map[vm_type]
        config["orca_config"]["num_devices"] = num_gpus

        config["scaler_config"] = {}
        config["scaler_config"]["min_replica_count"] = min_replicas
        config["scaler_config"]["max_replica_count"] = max_replicas

        request_data = {
            "project_id": str(project_id),
            "model_id": str(checkpoint_id),
            "deployment_type": deployment_type.value,
            "name": name,
            "vm": {"name": vm_type.value},
            "cloud": cloud.value,
            "region": region,
            "total_gpus": num_gpus,
            "infrequest_perm_check": security_level
            == DeploymentSecurityLevel.PROTECTED,
            "infrequest_log": logging,
            **config,
        }
        if description is not None:
            request_data["description"] = description

        client = DeploymentClient()
        deployment_raw = client.create_deployment(request_data)
        deployment = V1Deployment.model_validate(deployment_raw)
        return deployment

    @staticmethod
    def list(
        limit: int = 20,
        include_terminated: bool = False,
        from_oldest: bool = False,
        all_in_org: bool = False,
    ) -> List[V1Deployment]:
        """Lists deployments.

        Args:
            limit (int, optional): The maximum number of retrieved results. Defaults to 20.
            include_terminated (bool, optional): When True, includes the terminated deployments in the results. Defaults to False.
            from_oldest (bool, optional): List from the oldest deployment. Defaults to False.

        Raises:
            AuthenticationError: Raised when project context is not configured.

        Returns:
            List[V1Deployment]: Retrieved deployments.

        """
        if all_in_org:
            prj_id = None
        else:
            project_id = get_current_project_id()
            if project_id is None:
                raise AuthenticationError(
                    "Project context is not configured. "
                    "Set the project context either by 'pf.init' or 'pf project switch'."
                )
            prj_id = str(project_id)

        client = DeploymentClient()
        deployments = [
            V1Deployment.model_validate(deployment_raw)
            for deployment_raw in client.list_deployments(
                project_id=prj_id,
                archived=False,
                limit=limit,
                from_oldest=from_oldest,
            )
        ]
        num_active_deployments = len(deployments)
        if include_terminated and limit > num_active_deployments:
            deployments += [
                V1Deployment.model_validate(deployment_raw)
                for deployment_raw in client.list_deployments(
                    project_id=prj_id,
                    archived=True,
                    limit=limit - num_active_deployments,
                    from_oldest=from_oldest,
                )
            ]

        return deployments

    @staticmethod
    def get(id: str, *args, **kwargs) -> V1Deployment:
        """Gets deployment info.

        Args:
            id (str): ID of deployment to retrieve.

        Returns:
            V1Deployment: Retrieved deployment object.

        """
        client = DeploymentClient()
        deployment_raw = client.get_deployment(id)
        deployment = V1Deployment.model_validate(deployment_raw)

        return deployment

    @staticmethod
    def stop(id: str) -> None:
        """Stops a running deployment.

        Args:
            id (str): ID of deployment to stop.

        """
        client = DeploymentClient()
        client.stop_deployment(id)

    @staticmethod
    def get_metrics(id: str, time_window: int = 60) -> Dict[str, Any]:
        """Gets metrics of a deployment.

        Args:
            id (str): ID of deployment to get metrics.
            time_window (int, optional): Time window of results in seconds. Defaults to 60.

        Returns:
            Dict[str, Any]: Retrieved metrics data.

        """
        metrics_client = DeploymentMetricsClient(deployment_id=id)
        return metrics_client.get_metrics(deployment_id=id, time_window=time_window)

    @staticmethod
    def get_usage(since: datetime, until: datetime) -> Dict[str, Any]:
        """Gets usage info of a deployment.

        Args:
            since (datetime): Start datetime of the deployment usages to fetch.
            until (datetime): End datetime of the deployment usages to fetch.

        Returns:
            Dict[str, Any]: Retrieved deployment usage info.

        """
        client = PFSProjectUsageClient()
        return client.get_usage(since, until)

    @staticmethod
    def get_logs(id: str, replica_index: int = 0) -> List[Dict[str, Any]]:
        """Gets logs from a deployment.

        Args:
            id (str): ID of deployment to get logs.
            replica_index (int, optional): Index of deployment replica to retrieve logs. Defaults to 0.

        Returns:
            List[Dict[str, Any]]: Retrieved log data.

        """
        client = DeploymentLogClient(deployment_id=id)
        return client.get_deployment_logs(deployment_id=id, replica_index=replica_index)

    @staticmethod
    def adjust_replica_config(id: str, min_replicas: int, max_replicas: int) -> None:
        """Adjusts replica configuration of a running deployment.

        Args:
            id (str): ID of deployment to adjust the replica configuration.
            min_replicas (int): Minimum replica count.
            max_replicas (int): Maximum replica count.

        Raises:
            InvalidConfigError: Raised when any of the configurations (i.e., `min_replicas`, `max_replicas`) is invalid.

        """
        if min_replicas > max_replicas:
            raise InvalidConfigError(
                f"Should be min_replicas('{min_replicas}') <= max_replicas('{max_replicas}')."
            )
        client = DeploymentClient()
        client.update_deployment_scaler(
            deployment_id=id,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )

    @staticmethod
    def get_events(id: str) -> List[Dict[str, Any]]:
        """Gets events from a deployment.

        Args:
            id (str): ID of deployment to get events.

        Returns:
            List[Dict[str, Any]]: Deployment event info.

        """
        client = DeploymentEventClient(deployment_id=id)
        return client.get_events(deployment_id=id)

    @staticmethod
    def download_req_resp_logs(
        id: str, since: datetime, until: datetime, save_dir: Optional[str] = None
    ) -> None:
        """Downloads a file that has request-response logs.

        Args:
            id (str): ID of deployment.
            since (datetime): Start time of request-response logs.
            until (datetime): End time of request-response logs.
            save_dir (Optional[str], optional): Path to save logs. Set to the current directory if it is `None`. Defaults to None.

        Raises:
            InvalidPathError: Raised when `save_dir` does not exist or is read-only.
            InvalidConfigError: Raised when `since` <= `until` is not satisfied.
            NotFoundError: Raised when the deployment request-response logs are not found for the given time range.

        """
        if save_dir is not None and not os.path.isdir(save_dir):
            raise InvalidPathError(f"Directory '{save_dir}' is not found.")
        save_dir = save_dir or os.getcwd()

        if not os.access(save_dir, os.W_OK):
            raise InvalidPathError(f"Cannot save logs to {save_dir} which is readonly.")

        if since > until:
            raise InvalidConfigError(
                "Time value of `since` should be earlier than the value of `until`."
            )

        client = DeploymentReqRespClient(deployment_id=id)
        download_infos = client.get_download_urls(
            deployment_id=id, start=since, end=until
        )
        if len(download_infos) == 0:
            raise NotFoundError(f"No log exists for the deployment '{id}'.")

        for i, download_info in enumerate(download_infos):
            logger.info("Downloading files %d/%d...", i + 1, len(download_infos))
            full_storage_path = download_info["path"]
            deployment_id_part = extract_deployment_id_part(full_storage_path)
            timestamp_part = extract_datetime_part(full_storage_path)
            filename = f"{deployment_id_part}_{timestamp_part}.log"
            download_file(
                url=download_info["url"], out=os.path.join(save_dir, filename)
            )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow V1 Completion Serving API Schemas."""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel

from periflow.enums import BeamSearchType


class TokenSequence(BaseModel):
    """Token sequence schema."""

    tokens: List[int]


class V1CompletionOptions(BaseModel):
    """V1 completion options schema."""

    stream: Optional[bool] = None  # Enable streaming mode.
    prompt: Optional[str] = None  # Input prompt.
    tokens: Optional[List[int]] = None  # Input token sequence.
    timeout_microseconds: Optional[int] = None  # Request timeout in ms.
    max_tokens: Optional[int] = None  # Maximum output token count.
    max_total_tokens: Optional[int] = None  # Maximum input + output token count.
    min_tokens: Optional[int] = None  # Minimum output token count.
    min_total_tokens: Optional[int] = None  # Minimum input + output token count.
    n: Optional[int] = None  # The number of output sequences.
    num_beams: Optional[int] = None  # The number of beams in the beam search.
    length_penalty: Optional[float] = None  # Length penalty in the beam search.
    early_stopping: Optional[bool] = None  # Enable early stopping.
    no_repeat_ngram: Optional[int] = None  # Make N-gram not appear from the output.
    encoder_no_repeat_ngram: Optional[
        int
    ] = None  # Make N-gram not appear from the encoder output.
    repetition_penalty: Optional[float] = None  # Repetition penalty.
    encoder_repetition_penalty: Optional[float] = None  # Encoder repetition penality.
    temperature: Optional[float] = None  # Temperature.
    top_k: Optional[int] = None  # Top K.
    top_p: Optional[float] = None  # Top P.
    stop: Optional[List[str]] = None  # List of stop words.
    stop_tokens: Optional[List[TokenSequence]] = None  # List of stop tokens.
    seed: Optional[List[int]] = None  # Seed.
    beam_search_type: Optional[BeamSearchType] = None  # Beam search type.
    beam_compat_pre_normalization: Optional[bool] = None
    beam_compat_no_post_normalization: Optional[bool] = None
    bad_words: Optional[List[str]] = None  # List of bad words.
    bad_word_tokens: Optional[List[TokenSequence]] = None  # List of bad word tokens.
    include_output_logits: Optional[bool] = None  # Include logits in the output.
    eos_token: Optional[List[int]] = None  # List of EOS tokens.


class V1CompletionChoice(BaseModel):
    """V1 completion choice schema."""

    index: int
    seed: int
    text: str
    tokens: List[int]


class V1Completion(BaseModel):
    """V1 completion schema."""

    choices: List[V1CompletionChoice]


class V1CompletionLine(BaseModel):
    """V1 completion line schema."""

    event: str
    index: int
    text: str
    token: int



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Checkpoint Converter Interface."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict

import h5py  # type: ignore[import]
import torch
from tqdm.std import tqdm as std_tqdm

ENCODER_PREFIX = "encoder"
DECODER_PREFIX = "decoder"


class ConversionInterface(ABC):
    """Interface get information for converting common layers."""

    @property
    @abstractmethod
    def non_transformer_convert_dict(self) -> Dict[str, Any]:
        """Return the convert_dict for layers that do not belong to Transformer layers."""

    def convert_non_transformer_layers(
        self, state_dict: Dict[str, torch.Tensor], out_f: h5py.File, pbar: std_tqdm
    ) -> None:
        """Return the number of transformer layers converted in the decoder."""
        for converted_layer_name, (
            convert_fn,
            per_layer_postfixes,
        ) in self.non_transformer_convert_dict.items():
            converted_tensor = convert_fn(state_dict, "", per_layer_postfixes)
            out_f[converted_layer_name] = converted_tensor
            pbar.update()


class DecoderConversionInterface(ABC):
    """Mixin class get information for converting decoder layers."""

    @property
    @abstractmethod
    def decoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the decoder's transformer layer number."""

    @property
    @abstractmethod
    def decoder_layer_num(self) -> int:
        """Return the number of transformer layers in the decoder."""

    @property
    @abstractmethod
    def decoder_hidden_size(self) -> int:
        """Return the hidden size of the decoder."""

    @property
    @abstractmethod
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads."""

    @property
    @abstractmethod
    def decoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the decoder."""

    @property
    @abstractmethod
    def decoder_convert_dict(self) -> Dict[str, Any]:
        """Return the convert_dict for transformer layers in the decoder."""

    @property
    @abstractmethod
    def decoder_head_size(self) -> int:
        """Return the head size of the decoder."""

    def convert_decoder_layers(
        self, state_dict: Dict[str, torch.Tensor], out_f: h5py.File, pbar: std_tqdm
    ) -> None:
        """Return the number of transformer layers converted in the decoder."""
        out_f.create_group(DECODER_PREFIX)
        for i in range(self.decoder_layer_num):
            layer = self.decoder_layer_prefix + f"{i}"
            per_layer_out_ckpt = out_f[DECODER_PREFIX].create_group(f"h_._{i}")

            for converted_layer_name, (
                convert_fn,
                per_layer_postfixes,
            ) in self.decoder_convert_dict.items():
                converted_tensor = convert_fn(state_dict, layer, per_layer_postfixes)
                per_layer_out_ckpt[converted_layer_name] = converted_tensor
                pbar.update()


class EncoderConversionInterface(ABC):
    """Mixin class get information for converting encoder layers."""

    @property
    @abstractmethod
    def encoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the encoder's transformer layer number."""

    @property
    @abstractmethod
    def encoder_layer_num(self) -> int:
        """Return the number of transformer layers in the encoder."""

    @property
    @abstractmethod
    def encoder_hidden_size(self) -> int:
        """Return the hidden size of the encoder."""

    @property
    @abstractmethod
    def encoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the encoder."""

    @property
    @abstractmethod
    def encoder_convert_dict(self) -> Dict[str, Any]:
        """Return the convert_dict for transformer layers in the encoder."""

    @property
    @abstractmethod
    def encoder_head_size(self) -> int:
        """Return the head size of the encoder."""

    def convert_encoder_layers(
        self, state_dict: Dict[str, torch.Tensor], out_f: h5py.File, pbar: std_tqdm
    ) -> None:
        """Return the number of transformer layers converted in the encoder."""
        out_f.create_group(ENCODER_PREFIX)
        for i in range(self.encoder_layer_num):
            layer = self.encoder_layer_prefix + f"{i}"
            per_layer_out_ckpt = out_f[ENCODER_PREFIX].create_group(f"h_._{i}")

            for converted_layer_name, (
                convert_fn,
                per_layer_postfixes,
            ) in self.encoder_convert_dict.items():
                converted_tensor = convert_fn(
                    state_dict=state_dict,
                    layer=layer,
                    per_layer_postfixes=per_layer_postfixes,
                )
                per_layer_out_ckpt[converted_layer_name] = converted_tensor
                pbar.update()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Client: CLI and SDK for the fastest generative AI serving."""

from __future__ import annotations

import os

from periflow.schema.api.v1.completion import V1CompletionOptions
from periflow.sdk.api.completion import Completion
from periflow.sdk.init import init
from periflow.sdk.resource.checkpoint import Checkpoint
from periflow.sdk.resource.credential import Credential
from periflow.sdk.resource.deployment import Deployment

api_key = os.environ.get("PERIFLOW_API_KEY")
org_id = os.environ.get("PERIFLOW_ORG_ID")
project_id = os.environ.get("PERIFLOW_PRJ_ID")


__all__ = [
    "api_key",
    "org_id",
    "project_id",
    "init",
    "Checkpoint",
    "Credential",
    "Deployment",
    "Completion",
    "V1CompletionOptions",
]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI File System Management Utilities."""

from __future__ import annotations

import os
import re
import zipfile
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
from functools import wraps
from io import BufferedReader
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

import pathspec  # type: ignore
import requests
import typer
from dateutil.tz import tzlocal
from requests import Request, Session
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from periflow.utils.format import secho_error_and_exit
from periflow.utils.request import DEFAULT_REQ_TIMEOUT

# The actual hard limit of a part size is 5 GiB, and we use 200 MiB part size.
# See https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html.
S3_MPU_PART_MAX_SIZE = 200 * 1024 * 1024  # 200 MiB
S3_UPLOAD_SIZE_LIMIT = 5 * 1024 * 1024 * 1024  # 5 GiB

periflow_directory = Path.home() / ".periflow"


def get_periflow_directory() -> Path:
    """Get a hidden path to PeriFlow config directory."""
    periflow_directory.mkdir(exist_ok=True)
    return periflow_directory


@contextmanager
def zip_dir(
    base_path: Path, target_files: List[Path], zip_path: Path
) -> Iterator[BufferedReader]:
    """Zip direcotry.

    Args:
        base_path (Path): Base directory path. The target files will be located at the
            path related to the base path when the zip file is unzipped.
        target_files (List[Path]): All target files to zip.
        zip_path (Path): Path to save zip file.

    Yields:
        BufferedReader: Opened zip file object.

    """
    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_STORED) as zip_file:
        for file in target_files:
            zip_file.write(file, file.relative_to(base_path.parent))
    typer.secho("Uploading workspace directory...", fg=typer.colors.MAGENTA)
    try:
        yield zip_path.open("rb")
    finally:
        zip_path.unlink()


def get_workspace_files(dir_path: Path) -> List[Path]:
    """Get workspace file paths.

    Args:
        dir_path (Path): Path to the workspace directory.

    Returns:
        List[Path]: A list of paths to the file inside the workspace.

    """
    ignore_file = dir_path / ".pfignore"
    all_files = set(x for x in dir_path.rglob("*") if x.is_file() and x != ignore_file)

    if not ignore_file.exists():
        return list(all_files)

    with open(ignore_file, "r", encoding="utf-8") as f:
        ignore_patterns = f.read()
    spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern, ignore_patterns.splitlines()
    )
    matched_files = set(dir_path / x for x in spec.match_tree_files(dir_path))  # type: ignore
    return list(all_files.difference(matched_files))


def storage_path_to_local_path(storage_path: str, source_path: Path) -> str:
    """Translate a cloud storage path to the local file system path."""
    return strip_storage_path_prefix(str(source_path / storage_path))


def get_file_info(storage_path: str, source_path: Path) -> Dict[str, Any]:
    """Get file metadata."""
    loacl_path = storage_path_to_local_path(storage_path, source_path)
    return {
        "name": os.path.basename(storage_path),
        "path": storage_path,
        "mtime": datetime.fromtimestamp(
            os.stat(loacl_path).st_mtime, tz=tzlocal()
        ).isoformat(),
        "size": os.stat(loacl_path).st_size,
    }


def get_content_size(url: str) -> int:
    """Get download content size."""
    response = requests.get(url, stream=True, timeout=DEFAULT_REQ_TIMEOUT)
    if response.status_code != 200:
        secho_error_and_exit("Failed to download (invalid url)")
    return int(response.headers["Content-Length"])


def download_range(url: str, start: int, end: int, output: str, ctx: tqdm) -> None:
    """Download a specific part of a file from the URL."""
    headers = {"Range": f"bytes={start}-{end}"}
    response = requests.get(
        url, headers=headers, stream=True, timeout=DEFAULT_REQ_TIMEOUT
    )

    with open(output, "wb") as f:
        wrapped_object = CallbackIOWrapper(ctx.update, f, "write")
        for part in response.iter_content(1024):
            wrapped_object.write(part)


def download_file_simple(url: str, out: str, content_length: int) -> None:
    """Download a file without parallelism."""
    response = requests.get(url, stream=True, timeout=DEFAULT_REQ_TIMEOUT)
    with tqdm.wrapattr(
        open(out, "wb"), "write", miniters=1, total=content_length
    ) as fout:
        for chunk in response.iter_content(chunk_size=4096):
            fout.write(chunk)


def download_file_parallel(
    url: str, out: str, content_length: int, chunk_size: int = 1024 * 1024 * 4
) -> None:
    """Download a file in parallel."""
    chunks = range(0, content_length, chunk_size)

    temp_out_prefix = os.path.join(os.path.dirname(out), f".{os.path.basename(out)}")

    try:
        with tqdm(
            total=content_length, unit="B", unit_scale=True, unit_divisor=1024
        ) as t:
            with ThreadPoolExecutor() as executor:
                futs = [
                    executor.submit(
                        download_range,
                        url,
                        start,
                        start + chunk_size - 1,
                        f"{temp_out_prefix}.part{i}",
                        t,
                    )
                    for i, start in enumerate(chunks)
                ]
                wait(futs, return_when=FIRST_EXCEPTION)

        # Merge partitioned files
        with open(out, "wb") as f:
            for i in range(len(chunks)):
                chunk_path = f"{temp_out_prefix}.part{i}"
                with open(chunk_path, "rb") as chunk_f:
                    f.write(chunk_f.read())

                os.remove(chunk_path)
    finally:
        # Clean up zombie temporary partitioned files
        for i in range(len(chunks)):
            chunk_path = f"{temp_out_prefix}.part{i}"
            if os.path.isfile(chunk_path):
                os.remove(chunk_path)


def download_file(url: str, out: str) -> None:
    """Download a file from the URL."""
    file_size = get_content_size(url)

    # Create directory if not exists
    dirpath = os.path.dirname(out)
    try:
        os.makedirs(dirpath, exist_ok=True)
    except OSError as exc:
        secho_error_and_exit(
            f"Cannot create directory({dirpath}) to download file: {exc!r}"
        )

    if file_size >= 16 * 1024 * 1024:
        download_file_parallel(url, out, file_size)
    else:
        download_file_simple(url, out, file_size)


class FileSizeType(Enum):
    """File size type."""

    LARGE = "LARGE"
    SMALL = "SMALL"


def _filter_large_files(paths: List[str]) -> List[str]:
    return [path for path in paths if get_file_size(path) >= S3_UPLOAD_SIZE_LIMIT]


def _filter_small_files(paths: List[str]) -> List[str]:
    res = []
    for path in paths:
        size = get_file_size(path)
        if size == 0:
            # NOTE: S3 does not support file uploading for 0B size files.
            typer.secho(
                f"Skip uploading file ({path}) with size 0B.", fg=typer.colors.RED
            )
            continue
        if size < S3_UPLOAD_SIZE_LIMIT:
            res.append(path)
    return res


def filter_files_by_size(paths: List[str], size_type: FileSizeType) -> List[str]:
    """Filter files by size.

    Args:
        paths (List[str]): List of paths.
        size_type (FileSizeType): File size type to filter.

    Returns:
        List[str]: Filtered file paths.

    """
    handler_map = {
        FileSizeType.LARGE: _filter_large_files,
        FileSizeType.SMALL: _filter_small_files,
    }
    return handler_map[size_type](paths)


def expand_paths(path: Path, size_type: FileSizeType) -> List[str]:
    """Expand all file paths from the source path.

    Args:
        path (Path): The source path to expand files
        size_type (FileSizeType): The file size type to filter

    Returns:
        List[str]: A list of file paths from the source path
    """
    if path.is_file():
        paths = [str(path)]
        paths = filter_files_by_size(paths, size_type)
    else:
        paths = [str(p) for p in path.rglob("*")]
        paths = filter_files_by_size(paths, size_type)

    # Filter files only
    paths = [path for path in paths if os.path.isfile(path)]

    return paths


def upload_file(file_path: str, url: str, ctx: tqdm) -> None:
    """Upload a file to the URL."""
    try:
        with open(file_path, "rb") as f:
            fileno = f.fileno()
            total_file_size = os.fstat(fileno).st_size
            if total_file_size == 0:
                typer.secho(
                    f"The file with 0B size ({file_path}) is skipped.",
                    fg=typer.colors.RED,
                )
                return

            wrapped_object = CallbackIOWrapper(ctx.update, f, "read")
            with Session() as s:
                req = Request("PUT", url, data=wrapped_object)
                prep = req.prepare()
                prep.headers["Content-Length"] = str(
                    total_file_size
                )  # necessary to use ``CallbackIOWrapper``
                response = s.send(prep)
            if response.status_code != 200:
                secho_error_and_exit(
                    f"Failed to upload file ({file_path}): {response.content!r}"
                )
    except FileNotFoundError:
        secho_error_and_exit(f"{file_path} is not found.")


def get_file_size(file_path: str, prefix: Optional[str] = None) -> int:
    """Calculate a file size in bytes.

    Args:
        file_path (str): Path to the target file.
        prefix (Optional[str], optional): Attach the prefix to the path.

    Returns:
        int: The size of a file.
    """
    if prefix is not None:
        file_path = os.path.join(prefix, file_path)

    return os.stat(file_path).st_size


def get_total_file_size(file_paths: List[str], prefix: Optional[str] = None) -> int:
    """Get total file size in the paths."""
    return sum(get_file_size(file_path, prefix) for file_path in file_paths)


class CustomCallbackIOWrapper(CallbackIOWrapper):
    """Cutom callback IO wrapper."""

    def __init__(self, callback, stream, method="read", chunk_size=None) -> None:
        """Initialize custom callback IO wrapper."""
        # Wrap a file-like object's `read` or `write` to report data length to the `callback`.
        super().__init__(callback, stream, method)
        self._chunk_size = chunk_size
        self._cursor = 0

        func = getattr(stream, method)
        if method == "write":

            @wraps(func)
            def write(data, *args, **kwargs):
                res = func(data, *args, **kwargs)
                callback(len(data))
                return res

            self.wrapper_setattr("write", write)
        elif method == "read":

            @wraps(func)
            def read(*args, **kwargs):
                assert chunk_size is not None
                if self._cursor >= chunk_size:
                    self._cursor = 0
                    return None

                data = func(*args, **kwargs)
                data_size = len(data)  # default to 8 KiB
                callback(data_size)
                self._cursor += data_size
                return data

            self.wrapper_setattr("read", read)
        else:
            raise KeyError("Can only wrap read/write methods")


# pylint: disable=too-many-locals
def upload_part(
    file_path: str,
    chunk_index: int,
    part_number: int,
    upload_url: str,
    ctx: tqdm,
    is_last_part: bool,
) -> Dict[str, Any]:
    """Upload a specific part of the multipart upload payload.

    Args:
        file_path (str): Path to file to upload
        chunk_index (int): Chunk index to upload
        part_number (int): Part number of the multipart upload
        upload_url (str): A presigned URL for the multipart upload
        ctx (tqdm): tqdm context to update the progress
        is_last_part (bool): Whether this part is the last part of the payload.

    Returns:
        Dict[str, Any]: _description_
    """
    with open(file_path, "rb") as f:
        fileno = f.fileno()
        total_file_size = os.fstat(fileno).st_size
        cursor = chunk_index * S3_MPU_PART_MAX_SIZE
        f.seek(cursor)
        chunk_size = min(S3_MPU_PART_MAX_SIZE, total_file_size - cursor)
        wrapped_object = CustomCallbackIOWrapper(ctx.update, f, "read", chunk_size)
        with Session() as s:
            req = Request("PUT", upload_url, data=wrapped_object)
            prep = req.prepare()
            prep.headers["Content-Length"] = str(chunk_size)
            response = s.send(prep)
        response.raise_for_status()

        if is_last_part:
            assert not f.read(
                S3_MPU_PART_MAX_SIZE
            ), "Some parts of your data is not uploaded. Please try again."

    etag = response.headers["ETag"]
    return {
        "etag": etag,
        "part_number": part_number,
    }


def strip_storage_path_prefix(path: str) -> str:
    """Strip a prefix about auxiliary checkpoint info from the path.

    Args:
        path (str): Actual checkpoint storage path. The path may starts with the prefix
                    that contain the iteration number and distributed training configuration

    Returns:
        str: Path without the prefix
    """
    return re.sub(
        pattern=r"iter_\d{7}/mp\d{3}-\d{3}pp\d{3}-\d{3}/",
        repl="",
        string=path,
    )


def attach_storage_path_prefix(
    path: str,
    iteration: int,
    mp_rank: int,
    mp_degree: int,
    pp_rank: int,
    pp_degree: int,
) -> str:
    """Attach a filename prefix to mark up iteration and dist info.

    Args:
        path (str): The relative path to file
        iteration (int): Checkpoint iteration number
        mp_rank (int): Model parallelism(a.k.a. tensor parallelism) rank
        mp_degree (int): Model parallelism(a.k.a. tensor parallelism) degree
        pp_rank (int): Pipelined model parallelism rank
        pp_degree (int): Pipelined model parallelism degree

    Returns:
        str: Storage path with the prefix

    """
    # pylint: disable=line-too-long
    return f"iter_{iteration:07d}/mp{mp_rank:03d}-{mp_degree:03d}pp{pp_rank:03d}-{pp_degree:03d}/{path}"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow VM CLI."""

from __future__ import annotations

import typer

from periflow.client.deployment import PFSVMClient
from periflow.enums import GpuType
from periflow.formatter import TableFormatter

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)
quota_app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
    deprecated=True,
)

serving_vm_formatter = TableFormatter(
    name="Serving VM instances",
    fields=[
        "cloud",
        "region",
        "gpu_type",
        "vm.name",
        "vm.total_gpus",
        "vm.vcpu",
        "memory",
    ],
    headers=[
        "Cloud",
        "Region",
        "GPU type",
        "VM type",
        "GPU",
        "vCPUs",
        "Memory (GiB)",
    ],
)


# pylint: disable=redefined-builtin
@app.command()
def list():
    """List up available VMs."""
    pfs_vm_client = PFSVMClient()
    response = pfs_vm_client.list_vms()

    vm_dict_list = [
        {
            "cloud": nodegroup_list_dict["cloud"].upper(),
            "region": nodegroup_list_dict["region"],
            "vm": nodegroup["vm"],
            "gpu_type": nodegroup["vm"]["gpu_type"].upper(),
            "memory": int(nodegroup["vm"]["cpu_memory"]),
        }
        for nodegroup_list_dict in response
        for nodegroup in nodegroup_list_dict["nodegroup_list"]
        if nodegroup["vm"]["gpu_type"] in [gpu_type.value for gpu_type in GpuType]
    ]
    serving_vm_formatter.render(vm_dict_list)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=redefined-builtin, too-many-locals, too-many-arguments

"""PeriFlow Checkpoint CLI."""

from __future__ import annotations

import os
from datetime import datetime
from typing import List, Optional
from uuid import UUID

import typer

from periflow.enums import CheckpointCategory, CheckpointDataType, StorageType
from periflow.errors import (
    CheckpointConversionError,
    InvalidAttributesError,
    InvalidConfigError,
    InvalidPathError,
    NotFoundError,
    NotSupportedError,
)
from periflow.formatter import (
    JSONFormatter,
    PanelFormatter,
    TableFormatter,
    TreeFormatter,
)
from periflow.schema.resource.v1.checkpoint import V1Checkpoint
from periflow.sdk.resource.checkpoint import Checkpoint as CheckpointAPI
from periflow.utils.format import datetime_to_pretty_str, secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

table_formatter = TableFormatter(
    name="Checkpoints",
    fields=[
        "id",
        "name",
        "model_category",
        "iteration",
        "created_at",
    ],
    headers=[
        "ID",
        "Name",
        "Source",
        "Iteration",
        "Created At",
    ],
)
panel_formatter = PanelFormatter(
    name="Overview",
    fields=[
        "id",
        "name",
        "model_category",
        "forms[0].vendor",
        "forms[0].storage_name",
        "iteration",
        "forms[0].form_category",
        "created_at",
        "status",
    ],
    headers=[
        "ID",
        "Name",
        "Source",
        "Cloud",
        "Storage Name",
        "Iteration",
        "Format",
        "Created At",
        "Status",
    ],
)
json_formatter = JSONFormatter(name="Attributes")
tree_formatter = TreeFormatter(name="Files")


def get_translated_checkpoint_status(ckpt: V1Checkpoint) -> str:
    """Gets translated checkpoint status from the checkpoint info."""
    if ckpt.hard_deleted:
        hard_deleted_at = (
            datetime.strftime(ckpt.hard_deleted_at, "%Y-%m-%d %H:%M:%S")
            if ckpt.hard_deleted_at
            else "N/A"
        )
        typer.secho(
            f"This checkpoint was hard-deleted at {hard_deleted_at}. "
            "You cannot use this checkpoint.",
            fg=typer.colors.RED,
        )
        return "[bold red]Hard-Deleted"
    if ckpt.deleted:
        deleted_at = (
            datetime.strftime(ckpt.deleted_at, "%Y-%m-%d %H:%M:%S")
            if ckpt.deleted_at
            else "N/A"
        )
        typer.secho(
            f"This checkpoint was deleted at {deleted_at}. "
            "Please restore it if you want use this.",
            fg=typer.colors.YELLOW,
        )
        return "[bold yellow]Soft-Deleted"
    return "[bold green]Active"


@app.command()
def list(
    source: Optional[CheckpointCategory] = typer.Option(
        None,
        "--source",
        "-s",
        help="Source of checkpoints.",
    ),
    limit: int = typer.Option(
        20,
        "--limit",
        "-l",
        help="The number of recent checkpoints to see.",
    ),
    deleted: bool = typer.Option(
        False, "--deleted", "-d", help="Shows deleted checkpoint."
    ),
):
    """Lists all checkpoints that belong to a user's organization."""
    checkpoints = CheckpointAPI.list(category=source, limit=limit, deleted=deleted)
    ckpt_dicts = []
    for ckpt in checkpoints:
        ckpt_dict = ckpt.model_dump()
        ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
        ckpt_dicts.append(ckpt_dict)

    table_formatter.render(ckpt_dicts)


@app.command()
def view(
    checkpoint_id: UUID = typer.Argument(
        ..., help="ID of checkpoint to inspect in detail."
    )
):
    """Shows details of a checkpoint."""
    ckpt = CheckpointAPI.get(id=checkpoint_id)
    ckpt_dict = ckpt.model_dump()

    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    status = get_translated_checkpoint_status(ckpt)
    ckpt_dict["status"] = status

    panel_formatter.render([ckpt_dict])

    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


@app.command()
def create(
    name: str = typer.Option(
        ..., "--name", "-n", help="Name of your checkpoint to create."
    ),
    credential_id: UUID = typer.Option(
        ...,
        "--credential-id",
        "-i",
        help="ID of crendential to access cloud storage.",
    ),
    cloud_storage: StorageType = typer.Option(
        ...,
        "--cloud-storage",
        "-c",
        help="The cloud storage vendor where the checkpoint is uploaded.",
    ),
    storage_name: str = typer.Option(
        ...,
        "--storage-name",
        "-s",
        help="The name of the cloud storage where the checkpoint is uploaded.",
    ),
    region: str = typer.Option(
        ...,
        "--region",
        "-r",
        help="The cloud storage region where the checkpoint is uploaded.",
    ),
    storage_path: Optional[str] = typer.Option(
        None,
        "--storage-path",
        "-p",
        help="File or directory path to cloud storage. Defaults to the root of the storage.",
    ),
    iteration: Optional[int] = typer.Option(
        None, "--iteration", help="The iteration number of the checkpoint."
    ),
    attr_file: Optional[str] = typer.Option(
        None,
        "--attr-file",
        "-f",
        help=(
            "Path to the file containing checkpoint attributes. The file should be in "
            "YAML format."
        ),
    ),
):
    """Creates a checkpoint by registering checkpoint files in user's cloud storage to PeriFlow.

    An example of attribute files is as follows.
    You have the flexibility to modify the value of each field according to your preferences,
    but it is mandatory to fill in every field.

    ```yaml
    # blenderbot
    model_type: blenderbot
    dtype: fp16
    head_size: 80
    num_heads: 32
    hidden_size: 2560
    ff_intermediate_size: 10240
    num_encoder_layers: 2
    num_decoder_layers: 24
    max_input_length: 128
    max_output_length: 128
    vocab_size: 8008
    eos_token: 2
    decoder_start_token: 1

    # bloom
    model_type: bloom
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 30
    max_length: 2048
    vocab_size: 250880
    eos_token: 2

    # gpt
    model_type: gpt
    dtype: fp16
    head_size: 64
    num_heads: 25
    num_layers: 48
    max_length: 1024
    vocab_size: 50257
    eos_token: 50256

    # gpt-j
    model_type: gpt-j
    dtype: fp16
    head_size: 256
    rotary_dim: 64
    num_heads: 16
    num_layers: 28
    max_length: 2048
    vocab_size: 50400
    eos_token: 50256

    # gpt-neox
    model_type: gpt-neox
    dtype: fp16
    head_size: 128
    rotary_dim: 32
    num_heads: 40
    num_layers: 36
    max_length: 2048
    vocab_size: 50280
    eos_token: 0

    # llama
    model_type: llama
    dtype: fp16
    head_size: 128
    rotary_dim: 128
    num_heads: 32
    num_kv_heads: 32
    num_layers: 32
    ff_intermediate_size: 11008
    max_length: 2048
    vocab_size: 32000
    eos_token: 1

    # opt
    model_type: opt
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 32
    max_length: 2048
    vocab_size: 50272
    eos_token: 2

    # t5
    model_type: t5
    dtype: fp16
    head_size: 128
    num_heads: 32
    hidden_size: 1024
    ff_intermediate_size: 16384
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32100
    eos_token: 1
    decoder_start_token: 0

    # t5-v1_1
    model_type: t5-v1_1
    dtype: fp16
    head_size: 64
    num_heads: 32
    hidden_size: 2048
    ff_intermediate_size: 5120
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32128
    eos_token: 1
    decoder_start_token: 0
    ```

    :::tip
    If you want to use tokenizer, `tokenizer.json` file should be in the same directory
    as checkpoint file(`*.h5`).
    :::

    """
    try:
        ckpt = CheckpointAPI.create(
            name=name,
            credential_id=credential_id,
            cloud_storage=cloud_storage,
            region=region,
            storage_name=storage_name,
            storage_path=storage_path,
            iteration=iteration,
            attr_file_path=attr_file,
        )
    except (InvalidConfigError, NotSupportedError, InvalidAttributesError) as exc:
        secho_error_and_exit(str(exc))

    ckpt_dict = ckpt.model_dump()
    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    get_translated_checkpoint_status(ckpt)

    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


@app.command()
def delete(
    checkpoint_ids: List[UUID] = typer.Argument(
        ...,
        help=(
            "IDs of checkpoint to delete. "
            "When multiple IDs are provided in a space-separated string format, all "
            "corresponding checkpoints will be deleted."
        ),
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Forcefully delete checkpoint without confirmation prompt.",
    ),
):
    """Deletes the existing checkpoint.

    :::info
    Deleting linked checkpoints created with `pf checkpoint create` simply unlinks the
    checkpoint from the PeriFlow system. That is, it does not physically delete the
    checkpoint objects from the source cloud storage (your cloud storage).

    Uploaded checkpoints are physically deleted.
    :::

    """
    targets_str = ""
    for checkpoint_id in checkpoint_ids:
        targets_str += str(checkpoint_id)
        targets_str += "\n"

    if not force:
        do_delete = typer.confirm(
            f"Following checkpoints will be deleted:\n\n{targets_str}\n"
            "Are your sure to delete these?"
        )
        if not do_delete:
            raise typer.Abort()

    for checkpoint_id in checkpoint_ids:
        CheckpointAPI.delete(id=checkpoint_id)

    typer.secho("Checkpoints are deleted successfully!", fg=typer.colors.BLUE)


@app.command()
def download(
    checkpoint_id: UUID = typer.Argument(..., help="ID of checkpoint to download."),
    save_directory: Optional[str] = typer.Option(
        None,
        "--output",
        "-o",
        help="Output path to the directory to save checkpoint files.",
    ),
):
    """Downloads checkpoint files to local storage.

    :::info
    You cannot download checkpoints linked from your cloud storage.
    :::

    """
    try:
        CheckpointAPI.download(id=checkpoint_id, save_dir=save_directory)
    except InvalidPathError as exc:
        secho_error_and_exit(str(exc))


@app.command()
def upload(
    name: str = typer.Option(
        ..., "--name", "-n", help="Name of the checkpoint to upload"
    ),
    source_path: str = typer.Option(
        ..., "--source-path", "-p", help="Path to source file or directory to upload"
    ),
    iteration: Optional[int] = typer.Option(
        None, "--iteration", help="The iteration number of the checkpoint."
    ),
    attr_file: Optional[str] = typer.Option(
        None,
        "--attr-file",
        "-f",
        help="Path to the file containing the checkpoint attributes. The file should "
        "be in YAML format.",
    ),
    max_workers: int = typer.Option(
        min(32, (os.cpu_count() or 1) + 4),  # default of ``ThreadPoolExecutor``
        "--max-workers",
        "-w",
        help="The number of threads to upload files.",
    ),
):
    """Creates a checkpoint by uploading local checkpoint files.

    An example of attribute files is as follows.
    You have the flexibility to modify the value of each field according to your preferences,
    but it is mandatory to fill in every field.

    ```yaml
    # blenderbot
    model_type: blenderbot
    dtype: fp16
    head_size: 80
    num_heads: 32
    hidden_size: 2560
    ff_intermediate_size: 10240
    num_encoder_layers: 2
    num_decoder_layers: 24
    max_input_length: 128
    max_output_length: 128
    vocab_size: 8008
    eos_token: 2
    decoder_start_token: 1

    # bloom
    model_type: bloom
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 30
    max_length: 2048
    vocab_size: 250880
    eos_token: 2

    # gpt
    model_type: gpt
    dtype: fp16
    head_size: 64
    num_heads: 25
    num_layers: 48
    max_length: 1024
    vocab_size: 50257
    eos_token: 50256

    # gpt-j
    model_type: gpt-j
    dtype: fp16
    head_size: 256
    rotary_dim: 64
    num_heads: 16
    num_layers: 28
    max_length: 2048
    vocab_size: 50400
    eos_token: 50256

    # gpt-neox
    model_type: gpt-neox
    dtype: fp16
    head_size: 128
    rotary_dim: 32
    num_heads: 40
    num_layers: 36
    max_length: 2048
    vocab_size: 50280
    eos_token: 0

    # llama
    model_type: llama
    dtype: fp16
    head_size: 128
    rotary_dim: 128
    num_heads: 32
    num_kv_heads: 32
    num_layers: 32
    ff_intermediate_size: 11008
    max_length: 2048
    vocab_size: 32000
    eos_token: 1

    # opt
    model_type: opt
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 32
    max_length: 2048
    vocab_size: 50272
    eos_token: 2

    # t5
    model_type: t5
    dtype: fp16
    head_size: 128
    num_heads: 32
    hidden_size: 1024
    ff_intermediate_size: 16384
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32100
    eos_token: 1
    decoder_start_token: 0

    # t5-v1_1
    model_type: t5-v1_1
    dtype: fp16
    head_size: 64
    num_heads: 32
    hidden_size: 2048
    ff_intermediate_size: 5120
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32128
    eos_token: 1
    decoder_start_token: 0
    ```

    :::tip
    If you want to use tokenizer, `tokenizer.json` file should be in the same directory
    as checkpoint file(`*.h5`).
    :::

    """
    try:
        ckpt = CheckpointAPI.upload(
            name=name,
            source_path=source_path,
            iteration=iteration,
            attr_file_path=attr_file,
            max_workers=max_workers,
        )
    except (InvalidPathError, InvalidConfigError, InvalidAttributesError) as exc:
        secho_error_and_exit(str(exc))

    ckpt_dict = ckpt.model_dump()
    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    ckpt_dict["status"] = get_translated_checkpoint_status(ckpt)
    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


@app.command()
def restore(
    checkpoint_id: UUID = typer.Argument(..., help="ID of checkpoint to restore.")
):
    """Restores deleted checkpoint (available within 24 hours from the deletion).

    When a uploaded checkpoint is deleted, it is not physically deleted right away.
    Instead, it is soft-deleted and can be restored within 24 hours after the deletion.
    After the 24-hour retention period, it is hard-deleted and cannot be restored.

    :::caution
    When you delete a linked checkpoint, you cannot restore it. Instead, relink the
    original checkpoint using `pf checkpoint create`.
    :::

    """
    try:
        CheckpointAPI.restore(id=checkpoint_id)
    except NotFoundError as exc:
        secho_error_and_exit(str(exc))

    typer.secho(f"Checkpoint({checkpoint_id}) is successfully restored.")


@app.command()
def convert(
    model_name_or_path: str = typer.Option(
        ...,
        "--model-name-or-path",
        "-m",
        help="Hugging Face pretrained model name or path to the saved model checkpoint.",
    ),
    output_dir: str = typer.Option(
        ...,
        "--output-dir",
        "-o",
        help=(
            "Directory path to save the converted checkpoint and related configuration "
            "files. Three files will be created in the directory: `model.h5`, "
            "`tokenizer.json`, and `attr.yaml`. "
            "The `model.h5` is the converted checkpoint and can be renamed using "
            "the `--output-model-filename` option. "
            "The `tokenizer.json` is the PeriFlow-compatible tokenizer file, which should "
            "be uploaded along with the checkpoint file to tokenize the model input "
            "and output. "
            "The `attr.yaml` is the checkpoint attribute file, to be used when uploading "
            "the converted model to PeriFlow. You can designate the file name using "
            "the `--output-attr-filename` option."
        ),
    ),
    data_type: CheckpointDataType = typer.Option(
        ..., "--data-type", "-dt", help="The data type of converted checkpoint."
    ),
    cache_dir: Optional[str] = typer.Option(
        None, "--cache-dir", help="Directory for downloading checkpoint."
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Only check conversion avaliability."
    ),
    output_model_file_name: str = typer.Option(
        "model.h5",
        "--output-model-filename",
        help="Name of the converted checkpoint file.",
    ),
    output_attr_file_name: str = typer.Option(
        "attr.yaml",
        "--output-attr-filename",
        help="Name of the checkpoint attribute file.",
    ),
):
    """Convert huggingface's model checkpoint to PeriFlow format.

    When a checkpoint is in the Hugging Face format, it cannot be directly served;
    rather, it requires conversion to the PeriFlow format for serving. The conversion
    process involves copying the original checkpoint and transforming it into a
    checkpoint in the PeriFlow format (*.h5).

    :::caution
    The `pf checkpoint convert` is available only when the package is installed with
    `pip install periflow-client[mllib]`.
    :::

    """
    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            secho_error_and_exit(f"'{output_dir}' exists, but its not a directory.")
        os.mkdir(output_dir)

    model_output_path = os.path.join(output_dir, output_model_file_name)
    tokenizer_output_dir = output_dir
    attr_output_path = os.path.join(output_dir, output_attr_file_name)

    try:
        CheckpointAPI.convert(
            model_name_or_path=model_name_or_path,
            model_output_path=model_output_path,
            data_type=data_type,
            tokenizer_output_dir=tokenizer_output_dir,
            attr_output_path=attr_output_path,
            cache_dir=cache_dir,
            dry_run=dry_run,
        )
    except (NotFoundError, CheckpointConversionError) as exc:
        secho_error_and_exit(str(exc))

    msg = (
        f"Checkpoint({model_name_or_path}) can be converted."
        if dry_run
        else f"Checkpoint({model_name_or_path}) is converted successfully."
    )
    typer.secho(msg)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Defining PeriFlow Checkpoint Converter maps."""

from __future__ import annotations

from typing import Dict, Tuple, Type, Union

from transformers import (  # type: ignore[import]
    AutoModelForCausalLM,
    BlenderbotForConditionalGeneration,
    BloomForCausalLM,
    CodeGenForCausalLM,
    FalconForCausalLM,
    GPT2LMHeadModel,
    GPTJForCausalLM,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    OPTForCausalLM,
    PreTrainedModel,
    T5ForConditionalGeneration,
)

from periflow.converter.base import AbstractConverter
from periflow.converter.models.blenderbot import BlenderbotConverter
from periflow.converter.models.bloom import BloomForCausalLMConverter
from periflow.converter.models.codegen import CodegenForCausalLMConverter
from periflow.converter.models.falcon import FalconForCausalLMConverter
from periflow.converter.models.gpt2 import GPT2LMHeadModelConverter
from periflow.converter.models.gpt_neox import GPTNeoXForCausalLMConverter
from periflow.converter.models.gptj import GPTJForCausalLMConverter
from periflow.converter.models.llama import LlamaForCausalLMConverter
from periflow.converter.models.mpt import MPTForCausalLMConverter
from periflow.converter.models.opt import OPTForCausalLMConverter
from periflow.converter.models.t5 import T5Converter

model_arch_converter_map: Dict[
    str, Tuple[Union[AutoModelForCausalLM, PreTrainedModel], Type[AbstractConverter]]
] = {
    "BlenderbotForConditionalGeneration": (
        BlenderbotForConditionalGeneration,
        BlenderbotConverter,
    ),
    "BloomForCausalLM": (BloomForCausalLM, BloomForCausalLMConverter),
    "CodeGenForCausalLM": (CodeGenForCausalLM, CodegenForCausalLMConverter),
    "FalconForCausalLM": (FalconForCausalLM, FalconForCausalLMConverter),
    "GPTNeoXForCausalLM": (GPTNeoXForCausalLM, GPTNeoXForCausalLMConverter),
    "GPT2LMHeadModel": (GPT2LMHeadModel, GPT2LMHeadModelConverter),
    "GPTJForCausalLM": (GPTJForCausalLM, GPTJForCausalLMConverter),
    "LlamaForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "LLaMAForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "MPTForCausalLM": (AutoModelForCausalLM, MPTForCausalLMConverter),
    "OPTForCausalLM": (OPTForCausalLM, OPTForCausalLMConverter),
    "T5ForConditionalGeneration": (T5ForConditionalGeneration, T5Converter),
}
